/* FILE: demag.h            -*-Mode: c++-*-
 *
 * Average H demag field across rectangular cells.  This is a modified
 * version of the simpledemag class, which uses symmetries in the
 * interaction coefficients to reduce memory usage.
 *
 * This code uses the Oxs_FFT3v classes to perform direct FFTs of the
 * import magnetization ThreeVectors.  This Oxs_Demag class is a
 * drop-in replacement for an older Oxs_Demag class that used the
 * scalar Oxs_FFT class.  That older class has been renamed
 * Oxs_DemagOld, and is contained in the demagold.* files.
 *
 */

#ifndef _OXS_DEMAG
#define _OXS_DEMAG

#include "energy.h"
#include "fft3v.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"

/* End includes */

class Oxs_Demag:public Oxs_Energy {
private:
#if REPORT_TIME
  mutable Nb_StopWatch fftforwardtime;
  mutable Nb_StopWatch fftinversetime;
  mutable Nb_StopWatch convtime;
  mutable Nb_StopWatch dottime;
#endif // REPORT_TIME

  mutable INT4m rdimx; // Natural size of real data
  mutable INT4m rdimy; // Digital Mars compiler wants these as separate
  mutable INT4m rdimz; //    statements, because of "mutable" keyword.
  mutable INT4m cdimx; // Full size of complex data
  mutable INT4m cdimy;
  mutable INT4m cdimz;
  // 2*(cdimx-1)>=rdimx, cdimy>=rdimy, cdimz>=rdimz
  // cdim-1 and cdim[yz] should be powers of 2.
  mutable INT4m adimx; // Dimensions of A## storage (see below).
  mutable INT4m adimy;
  mutable INT4m adimz;

  mutable UINT4m mesh_id;

  // The A## arrays hold demag coefficients, transformed into
  // frequency domain.  These are held long term.  These are
  // held as arrays of three-vectors, for consistency with the
  // handling of the imported magnetization and exported field
  // arrays.  Due to symmetries, only the first octant needs
  // to be saved.
  //   The Hxfrm array is temporary space used primarily to hold
  // the transform image of the computed field.  It is also
  // used as intermediary storage for the A## values, which
  // for convenience are computed at full size before being
  // transfered to the 1/8-sized A## storage arrays.
  //   All of these arrays are actually arrays of complex-valued
  // three vectors, but are handled as simple REAL arrays.
  struct A_coefs {
    OXS_FFT_REAL_TYPE A00;
    OXS_FFT_REAL_TYPE A01;
    OXS_FFT_REAL_TYPE A02;
    OXS_FFT_REAL_TYPE A11;
    OXS_FFT_REAL_TYPE A12;
    OXS_FFT_REAL_TYPE A22;
  };
  mutable A_coefs* A;
  mutable OXS_FFT_REAL_TYPE *Hxfrm;

  REAL8m dipole_radius;
  /// If >=0, then radius beyond which demag coefficients A_coefs
  /// are computed using dipolar approximation instead of Newell's
  /// analytic formulae.

  mutable OXS_FFT_REAL_TYPE *Mtemp;  // Temporary space to hold
  /// Ms[]*m[].  The plan is to make this space unnecessary
  /// by introducing FFT functions that can take Ms as input
  /// and do the multiplication on the fly.

  // Object to perform FFTs.  All transforms are the same size, so we
  // only need one Oxs_FFT3DThreeVector object.  (Note: A
  // multi-threaded version of this code might want to have a separate
  // Oxs_FFT3DThreeVector for each thread, so that the roots-of-unity
  // arrays in each thread are independent.  This may improve memory
  // access on multi-processor machines where each processor has its
  // own memory.)
   //   The embed_convolution boolean specifies whether or not to embed
  // the convolution (matrix-vector multiply in FFT transform space)
  // computation inside the fftz computation.  If true then
  // embed_block_size is the number of z-axis FFTs to perform about
  // each "convolution" (really matrix-vector multiply) computation.
  // These are set inside the FillCoefficientArrays member function,
  // and used inside GetEnergy.
  mutable Oxs_FFT1DThreeVector fftx;
  mutable Oxs_FFTStrided ffty;
  mutable Oxs_FFTStrided fftz;
  mutable BOOL embed_convolution;
  mutable INT4m embed_block_size;
  INT4m cache_size; // Cache size in bytes.  Used to select
		    // embed_block_size.

  INT4m zero_self_demag;
  /// If zero_self_demag is true, then diag(1/3,1/3,1/3) is subtracted
  /// from the self-demag term.  In particular, for cubic cells this
  /// makes the self-demag field zero.  This will change the value
  /// computed for the demag energy by a constant amount, but since the
  /// demag field is changed by a multiple of m, the torque and
  /// therefore the magnetization dynamics are unaffected.

  void FillCoefficientArrays(const Oxs_Mesh* mesh) const;

  void ReleaseMemory() const;


protected:
  virtual void GetEnergy(const Oxs_SimState& state,
			 Oxs_EnergyData& oed) const;

public:
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.
  Oxs_Demag(const char* name,     // Child instance id
	    Oxs_Director* newdtr, // App director
	    const char* argstr);  // MIF input block parameters
  virtual ~Oxs_Demag();
  virtual BOOL Init();
};


#endif // _OXS_DEMAG
