/* FILE: edgecorranisotropy.cc            -*-Mode: c++-*-
 *
 * Edge Correction Anisotropy, derived from Oxs_Energy class.
 *
 */

#include <assert.h>
#include <string>

#include "oc.h"
#include "nb.h"
#include "vf.h"

#include "arrayscalarfield.h"
#include "director.h"
#include "threevector.h"
#include "simstate.h"
#include "ext.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "uniformvectorfield.h"
#include "util.h"
#include "rectangularmesh.h"
#include "edgecorranisotropy.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

OC_USE_STRING;

#if 0 && !defined(NDEBUG)
# define DEBUG_PRINT
#endif // NDEBUG

#ifdef DEBUG_PRINT
# include <stdio.h>
# include <string.h>
#endif

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_EdgeCorrAnisotropy);

/* End includes */

// Debug routine
#ifdef DEBUG_PRINT
void PrintMeshValue_Real
(
 const char* msg,const char* fmt,
 const Oxs_RectangularMesh* mesh,
 const Oxs_MeshValue<REAL8m>& val
)
{
  const INT4m xdim  = static_cast<INT4m>(mesh->DimX());
  const INT4m ydim  = static_cast<INT4m>(mesh->DimY());
  const INT4m zdim  = static_cast<INT4m>(mesh->DimZ());

  INT4m i,j,k;

  char eqline[] = "==================================="
    "=============================================";

  int msglen = strlen(msg);
  int prelen = (77 - msglen)/2;
  int postlen = 77 - prelen - msglen;
  if(prelen<1) prelen = 1;
  if(postlen<1) postlen = 1;

  printf("%1.*s %1.*s %1.*s\n",prelen,eqline,msglen,msg,postlen,eqline);
  for(k=0;k<zdim;++k) {
    printf("k=%d\n",k);
    for(j=ydim-1;j>=0;--j) {
      for(i=0;i<xdim;++i) {
        printf(fmt,val[mesh->Index(i,j,k)]);
      }
      printf("\n");
    }
    if(k+1<zdim) {
      printf(" --------------------------------------------------------\n");
    }      
  }
  printf("%79.79s\n",eqline);
}
void PrintMeshValue_int
(
 const char* msg,const char* fmt,
 const Oxs_RectangularMesh* mesh,
 const Oxs_MeshValue<int>& val
)
{
  const INT4m xdim  = static_cast<INT4m>(mesh->DimX());
  const INT4m ydim  = static_cast<INT4m>(mesh->DimY());
  const INT4m zdim  = static_cast<INT4m>(mesh->DimZ());

  INT4m i,j,k;

  char eqline[] = "==================================="
    "=============================================";

  int msglen = strlen(msg);
  int prelen = (77 - msglen)/2;
  int postlen = 77 - prelen - msglen;
  if(prelen<1) prelen = 1;
  if(postlen<1) postlen = 1;

  printf("%1.*s %1.*s %1.*s\n",prelen,eqline,msglen,msg,postlen,eqline);
  for(k=0;k<zdim;++k) {
    printf("k=%d\n",k);
    for(j=ydim-1;j>=0;--j) {
      for(i=0;i<xdim;++i) {
        printf(fmt,val[mesh->Index(i,j,k)]);
      }
      printf("\n");
    }
    if(k+1<zdim) {
      printf(" --------------------------------------------------------\n");
    }      
  }
  printf("%79.79s\n",eqline);
}
#endif // DEBUG_PRINT

// Helper method that computes Ms_coarse
void Oxs_EdgeCorrAnisotropy::ComputeMsCoarse
( // Imports
 Oxs_Director* director, // App director
 const INT4m (& subcount)[3], // Fine cells counts inside one coarse cell
 const Oxs_RectangularMesh* cmesh, // Coarse mesh
 const Oxs_ScalarField* Ms_fine_init, // Ms initializer for fine mesh
 // Exports
 Oxs_MeshValue<REAL8m>& Ms_coarse // Averaged Ms, on coarse grid
 )
{
  // Create small submesh to represent an individual cell in the
  // coarse mesh.  The submesh basept can be moved around by calls to
  // submesh->TranslateBasePt.  The translation_offset value is the
  // offset between the center of a coarse mesh cell and the basept
  // of the submesh when the submesh is nested inside the coarse
  // cell.
  ThreeVector coarse_cellsize(cmesh->EdgeLengthX(),
                              cmesh->EdgeLengthY(),
                              cmesh->EdgeLengthZ());
  ThreeVector sub_cellsize(coarse_cellsize.x/subcount[0],
                           coarse_cellsize.y/subcount[1],
                           coarse_cellsize.z/subcount[2]);
  ThreeVector translation_offset = 0.5*(sub_cellsize-coarse_cellsize);
    

  Oxs_Box sub_range;
  sub_range.Set(0.0,coarse_cellsize.x,
                0.0,coarse_cellsize.y,
                0.0,coarse_cellsize.z);
  Oxs_RectangularMesh submesh("Submesh",director,"",
                              sub_cellsize,sub_range);

  // Compute averaged Ms values for coarse mesh
  Ms_coarse.AdjustSize(cmesh);
  Oxs_MeshValue<REAL8m> Ms_fine; // Ms on fine mesh
  ThreeVector basept;
  const UINT4m coarse_count = cmesh->Size();
  const UINT4m fine_count = submesh.Size();
  const REAL8m cellscale = 1.0/static_cast<REAL8m>(fine_count);
  for(UINT4m i=0;i<coarse_count;++i) {
    cmesh->Center(i,basept);
    basept += translation_offset;
    submesh.TranslateBasePt(basept);  // Translate fine mesh
    Ms_fine_init->FillMeshValue(&submesh,Ms_fine); // Fill fine mesh

    // Find average value of Ms on fine mesh
    REAL8m sum = 0.0;
    for(UINT4m fi=0;fi<fine_count;++fi) sum += Ms_fine[fi];

    // Store Ms_ave
    Ms_coarse[i] = sum * cellscale;
  }

}

// Helper method that computes Ms_coarse and anisotropy corrections
void Oxs_EdgeCorrAnisotropy::ComputeAnisotropyCorrections
( // Imports
 Oxs_Director* director,     // Application director
 const INT4m (&subcount)[3],   // Fine cells counts inside one coarse cell
 const INT4m (&supercount)[3], // Coarse cell counts used in correction
 Oxs_Energy* cDemag,           // Demag object for coarse working mesh
 Oxs_Energy* fDemag,           // Demag object for fine working mesh
 const Oxs_RectangularMesh* cmesh, // Coarse mesh
 const Oxs_ScalarField* Msinit,    // Ms initializer
 // Exports
 Oxs_MeshValue<REAL8m>& Ms_coarse, // Averaged Ms, on coarse grid
 Oxs_MeshValue<ThreeVector>& Ncorr_diag,    // Correction terms; see
 Oxs_MeshValue<ThreeVector>& Ncorr_offdiag  // header file for details.
 )
{
  // NOTE: Unless you want total demag coefficient thrashing on the
  // demag objects for the working meshes, cDemag and fDemag had better
  // point to distinct objects.
  assert(cDemag != fDemag);

  INT4m i,j,k;

  // Coarse mesh dimensions
  const INT4m cxdim  = static_cast<INT4m>(cmesh->DimX());
  const INT4m cydim  = static_cast<INT4m>(cmesh->DimY());
  const INT4m czdim  = static_cast<INT4m>(cmesh->DimZ());
  const UINT4m cxydim 
    = static_cast<UINT4m>(cxdim) * static_cast<UINT4m>(cydim);

  // Create fine mesh
  Oxs_OwnedPointer<Oxs_RectangularMesh> fmesh;
  cmesh->MakeRefinedMesh("Oxs_EdgeCorrAnistropy_FineMesh",
                         subcount[0],subcount[1],subcount[2],fmesh);
  const INT4m fxdim  = static_cast<INT4m>(fmesh->DimX());
  const INT4m fydim  = static_cast<INT4m>(fmesh->DimY());
  const INT4m fzdim  = static_cast<INT4m>(fmesh->DimZ());
  const UINT4m fxydim
    = static_cast<UINT4m>(fxdim) * static_cast<UINT4m>(fydim);

  // Ms on fine mesh
  Oxs_MeshValue<REAL8m> Ms_fine;
  Msinit->FillMeshValue(fmesh.GetPtr(),Ms_fine);

#ifdef DEBUG_PRINT
  PrintMeshValue_Real("Ms_fine"," %1.0f",fmesh.GetPtr(),Ms_fine);
#endif // DEBUG_PRINT

  // Boolean array to keep track of which coarse cells are not
  // uniformly filled, and hence require correction.
  Oxs_MeshValue<int> cell_needs_correction;
  cell_needs_correction.AdjustSize(cmesh);

  // Compute averaged Ms values for coarse mesh
  Ms_coarse.AdjustSize(cmesh);
  const REAL8m cellscale
    = 1.0/static_cast<REAL8m>(subcount[0]*subcount[1]*subcount[2]);
  INT4m ic,jc,kc;
  for(kc=0;kc<czdim;++kc) {
    INT4m koff = kc*subcount[2];
    for(jc=0;jc<cydim;++jc) {
      INT4m joff = jc*subcount[1];
      for(ic=0;ic<cxdim;++ic) {
        INT4m ioff = ic*subcount[0];
        REAL8m sum = 0.0;
	REAL8m corrcheckval = Ms_fine[fmesh->Index(ioff,joff,koff)];
	int ncorr = 0; // Needs correction check; correction
	// is not needed if all cells in block have same Ms
        for(k=0;k<subcount[2];++k) {
          for(j=0;j<subcount[1];++j) {
            for(i=0;i<subcount[0];++i) {
	      REAL8m Msfval = Ms_fine[fmesh->Index(ioff+i,joff+j,koff+k)];
              sum += Msfval;
	      if(fabs(Msfval-corrcheckval)
		 >16*REAL8_EPSILON*(fabs(Msfval)+fabs(corrcheckval))) {
		ncorr = 1;
	      }
            }
          }
        }
        UINT4m cindex = cmesh->Index(ic,jc,kc);
        if(ncorr) {
          cell_needs_correction[cindex] = 1;
          REAL8m val = sum * cellscale;
          Ms_coarse[cindex] = val;
        } else {
          cell_needs_correction[cindex] = 0;
          Ms_coarse[cindex] = corrcheckval;
        }
      }
    }
  }

#ifdef DEBUG_PRINT
  PrintMeshValue_Real("Ms_coarse"," %4.2f",cmesh,Ms_coarse);
  PrintMeshValue_int("cell_needs_correction"," %4d",cmesh,
                     cell_needs_correction);
#endif // DEBUG_PRINT

  // Create appropriately sized working meshes
  Oxs_ThreeVector coarse_cellsize;
  coarse_cellsize.x = cmesh->EdgeLengthX();
  coarse_cellsize.y = cmesh->EdgeLengthY();
  coarse_cellsize.z = cmesh->EdgeLengthZ();

  Oxs_Box range;
  range.Set(0,coarse_cellsize.x*supercount[0],
            0,coarse_cellsize.y*supercount[1],
            0,coarse_cellsize.z*supercount[2]);


  Oxs_OwnedPointer<Oxs_RectangularMesh> cwmesh;
  cwmesh.SetAsOwner(new Oxs_RectangularMesh("EdgeCorrMeshCoarse",director,
                                            "",coarse_cellsize,range));
  Oxs_OwnedPointer<Oxs_RectangularMesh> fwmesh;
  cwmesh->MakeRefinedMesh("EdgeCorrMeshFine",
                          subcount[0],subcount[1],subcount[2],fwmesh);
  INT4m fwxdim = static_cast<INT4m>(fwmesh->DimX());
  INT4m fwxydim = static_cast<INT4m>(fwmesh->DimY())*fwxdim;


  // Working Ms arrays, coarse and fine.
  Oxs_MeshValue<REAL8m> cwMs;  cwMs.AdjustSize(cwmesh.GetPtr());
  Oxs_MeshValue<REAL8m> fwMs;  fwMs.AdjustSize(fwmesh.GetPtr());

  // Uniformly aligned spin arrays, for convenience use with
  // state spin arrays.
  Oxs_MeshValue<ThreeVector> cwspin_x;  cwspin_x.AdjustSize(cwmesh.GetPtr());
  Oxs_MeshValue<ThreeVector> cwspin_y;  cwspin_y.AdjustSize(cwmesh.GetPtr());
  Oxs_MeshValue<ThreeVector> cwspin_z;  cwspin_z.AdjustSize(cwmesh.GetPtr());
  INT4m cwsize = static_cast<INT4m>(cwmesh->Size());
  for(i=0;i<cwsize;++i) {
    cwspin_x[i].Set(1.,0.,0.);
    cwspin_y[i].Set(0.,1.,0.);
    cwspin_z[i].Set(0.,0.,1.);
  }
  Oxs_MeshValue<ThreeVector> fwspin_x;  fwspin_x.AdjustSize(fwmesh.GetPtr());
  Oxs_MeshValue<ThreeVector> fwspin_y;  fwspin_y.AdjustSize(fwmesh.GetPtr());
  Oxs_MeshValue<ThreeVector> fwspin_z;  fwspin_z.AdjustSize(fwmesh.GetPtr());
  INT4m fwsize = static_cast<INT4m>(fwmesh->Size());
  for(i=0;i<fwsize;++i) {
    fwspin_x[i].Set(1.,0.,0.);
    fwspin_y[i].Set(0.,1.,0.);
    fwspin_z[i].Set(0.,0.,1.);
  }

  // Working states, coarse and fine
  Oxs_SimState cwstate;
  cwstate.mesh = cwmesh.GetPtr();
  cwstate.Ms = &cwMs;
  Oxs_MeshValue<REAL8m> cwenergy;     // Buffer space for oed object in
  Oxs_MeshValue<ThreeVector> cwHfield; // GetEnergyData call
  const UINT4m ccenter_windex   // Center index
    = cwmesh->Index(cwmesh->DimX()/2,cwmesh->DimY()/2,cwmesh->DimZ()/2);

  Oxs_SimState fwstate;
  fwstate.mesh = fwmesh.GetPtr();
  fwstate.Ms = &fwMs;
  Oxs_MeshValue<REAL8m> fwenergy;     // Buffer space for oed object in
  Oxs_MeshValue<ThreeVector> fwHfield; // GetEnergyData call
  const UINT4m fcenter_windex   // Center index; actually, first cell in
    = fwmesh->Index((cwmesh->DimX()/2)*subcount[0],  /// block corresponding
                    (cwmesh->DimY()/2)*subcount[1],  /// to center block in
                    (cwmesh->DimZ()/2)*subcount[2]); /// coarse mesh

  // Indexing through coarse grid, compute and fill Ncorr arrays
  Ncorr_diag.AdjustSize(cmesh);
  Ncorr_offdiag.AdjustSize(cmesh);
  UINT4m total_coarse_size = cmesh->Size();
  for(UINT4m cindex=0;cindex<total_coarse_size;++cindex) {
    //   Does coarse grid index need correction?

    if(Ms_coarse[cindex]==0.0) {
      Ncorr_diag[cindex].x
        = Ncorr_diag[cindex].y
        = Ncorr_diag[cindex].z
        = Ncorr_offdiag[cindex].x
        = Ncorr_offdiag[cindex].y
        = Ncorr_offdiag[cindex].z
        = 0.0;
      continue;
    }

    ThreeVector Hdiffx,Hdiffy,Hdiffz;

    // Yes.  Fill cwMs and fwMs
    // with appropriate subrange from Ms_fine and Ms_coarse
    UINT4m kcoarse = cindex/cxydim;
    UINT4m jcoarse = (cindex - kcoarse*cxydim)/cxdim;
    UINT4m icoarse = cindex - kcoarse*cxydim - jcoarse*cxdim;

    UINT4m kfine = kcoarse * subcount[2];
    UINT4m jfine = jcoarse * subcount[1];
    UINT4m ifine = icoarse * subcount[0];

    // Fill coarse submesh arrays
    INT4m iw,jw,kw,windex;
    INT4m kmin = -supercount[2]/2;
    INT4m kmax = kmin + supercount[2];
    INT4m jmin = -supercount[1]/2;
    INT4m jmax = jmin + supercount[1];
    INT4m imin = -supercount[0]/2;
    INT4m imax = imin + supercount[0];
    windex = 0;
    int docorr = 0;
    for(kw = kmin; kw<kmax; ++kw) {
      k = kcoarse + kw;
      for(jw = jmin; jw<jmax; ++jw) {
        j = jcoarse + jw;
        for(iw= imin; iw<imax; ++iw) {
          i = icoarse + iw;
          if(i<0 || i>=cxdim || j<0 || j>=cydim || k<0 || k>=czdim) {
            cwMs[windex] = 0.0;
          } else {
	    int tindex = i+j*cxdim+k*cxydim;
            cwMs[windex] = Ms_coarse[tindex];
	    if(cell_needs_correction[tindex]) docorr = 1;
          }
          ++windex;
        }
      }
    }
    if(!docorr) {
      // All coarse cells in working neighborhood are full;
      // set anisotropy correction to 0
      Ncorr_diag[cindex].x
        = Ncorr_diag[cindex].y
        = Ncorr_diag[cindex].z
        = Ncorr_offdiag[cindex].x
        = Ncorr_offdiag[cindex].y
        = Ncorr_offdiag[cindex].z
        = 0.0;
      continue;
    }

    // Fill fine submesh arrays
    kmin = -(supercount[2]/2)*subcount[2];
    kmax = kmin + supercount[2]*subcount[2];
    jmin = -(supercount[1]/2)*subcount[1];
    jmax = jmin + supercount[1]*subcount[1];
    imin = -(supercount[0]/2)*subcount[0];
    imax = imin + supercount[0]*subcount[0];
    windex = 0;
    for(kw = kmin; kw<kmax; ++kw) {
      k = kfine + kw;
      for(jw = jmin; jw<jmax; ++jw) {
        j = jfine + jw;
        for(iw= imin; iw<imax; ++iw) {
          i = ifine + iw;
          if(i<0 || i>=fxdim || j<0 || j>=fydim || k<0 || k>=fzdim) {
            fwMs[windex] = 0.0;
          } else {
            REAL8m tMs = Ms_fine[i+j*fxdim+k*fxydim];
            fwMs[windex] = tMs;
          }
          ++windex;
        }
      }
    }

#ifdef DEBUG_PRINT
    printf("+++ Coarse mesh index (%2d,%2d,%2d) +++\n",icoarse,jcoarse,kcoarse);
    PrintMeshValue_Real("cwMs"," %7.2f",cwmesh.GetPtr(),cwMs);
    PrintMeshValue_Real("fwMs"," %7.2f",fwmesh.GetPtr(),fwMs);
#endif // DEBUG_PRINT


    //     Fill (swap) cwstate and fwstate spin in x, y, and z directions
    //       Call coarse demag to compute H with current spin direction
    //       Call fine   demag to compute H with current spin direction
    //       Determine correction for coarse grid index

    // spins in x-direction
    {
      cwstate.spin.Swap(cwspin_x);  // Spins in work states point in
      fwstate.spin.Swap(fwspin_x);  // x direction

      Oxs_EnergyData cwoed(cwstate);
      cwoed.energy_buffer = &cwenergy;
      cwoed.field_buffer  = &cwHfield;

      Oxs_EnergyData fwoed(fwstate);
      fwoed.energy_buffer = &fwenergy;
      fwoed.field_buffer  = &fwHfield;

      cDemag->GetEnergyData(cwstate,cwoed);
      fDemag->GetEnergyData(fwstate,fwoed);

      cwstate.spin.Swap(cwspin_x);  // Restore x pointing spins back to
      fwstate.spin.Swap(fwspin_x);  // c/fwspin storage.

      // Coarse field is just field at center of coarse mesh
      ThreeVector Hcoarse = (*cwoed.field)[ccenter_windex];

      // Fine field is average field across fine cells coincident
      // with center coarse cell
      ThreeVector Hfine(0.,0.,0.);
      for(k=0;k<subcount[2];++k) {
        for(j=0;j<subcount[1];++j) {
          for(i=0;i<subcount[0];++i) {
            windex = fcenter_windex + k*fwxydim + j*fwxdim + i;
#if 0 && defined(DEBUG_PRINT)
            printf(" windex (%3d,%3d,%3d)->%6d  fwMs=%g, H=(%g,%g,%g)\n",
                   i,j,k,windex,fwMs[windex],
                   (*fwoed.field)[windex].x,(*fwoed.field)[windex].y,
                   (*fwoed.field)[windex].z);
#endif
            Hfine += fwMs[windex] * (*fwoed.field)[windex];
          }
        }
      }
      Hfine *= 1.0/(cwMs[ccenter_windex]*subcount[0]*subcount[1]*subcount[2]);
      Hdiffx = Hfine - Hcoarse;
    }

    // spins in y-direction
    {
      cwstate.spin.Swap(cwspin_y);  // Spins in work states point in
      fwstate.spin.Swap(fwspin_y);  // x direction

      Oxs_EnergyData cwoed(cwstate);
      cwoed.energy_buffer = &cwenergy;
      cwoed.field_buffer  = &cwHfield;

      Oxs_EnergyData fwoed(fwstate);
      fwoed.energy_buffer = &fwenergy;
      fwoed.field_buffer  = &fwHfield;

      cDemag->GetEnergyData(cwstate,cwoed);
      fDemag->GetEnergyData(fwstate,fwoed);

      cwstate.spin.Swap(cwspin_y);  // Restore x pointing spins back to
      fwstate.spin.Swap(fwspin_y);  // c/fwspin storage.

      // Coarse field is just field at center of coarse mesh
      ThreeVector Hcoarse = (*cwoed.field)[ccenter_windex];

      // Fine field is average field across fine cells coincident
      // with center coarse cell
      ThreeVector Hfine(0.,0.,0.);
      for(k=0;k<subcount[2];++k) {
        for(j=0;j<subcount[1];++j) {
          for(i=0;i<subcount[0];++i) {
            windex = fcenter_windex + k*fwxydim + j*fwxdim + i;
            Hfine += fwMs[windex] * (*fwoed.field)[windex];
          }
        }
      }
      Hfine *= 1.0/(cwMs[ccenter_windex]*subcount[0]*subcount[1]*subcount[2]);
      Hdiffy = Hfine - Hcoarse;
    }

    // spins in z-direction
    {
      cwstate.spin.Swap(cwspin_z);  // Spins in work states point in
      fwstate.spin.Swap(fwspin_z);  // x direction

      Oxs_EnergyData cwoed(cwstate);
      cwoed.energy_buffer = &cwenergy;
      cwoed.field_buffer  = &cwHfield;

      Oxs_EnergyData fwoed(fwstate);
      fwoed.energy_buffer = &fwenergy;
      fwoed.field_buffer  = &fwHfield;

      cDemag->GetEnergyData(cwstate,cwoed);
      fDemag->GetEnergyData(fwstate,fwoed);

      cwstate.spin.Swap(cwspin_z);  // Restore x pointing spins back to
      fwstate.spin.Swap(fwspin_z);  // c/fwspin storage.

      // Coarse field is just field at center of coarse mesh
      ThreeVector Hcoarse = (*cwoed.field)[ccenter_windex];

      // Fine field is average field across fine cells coincident
      // with center coarse cell
      ThreeVector Hfine(0.,0.,0.);
      for(k=0;k<subcount[2];++k) {
        for(j=0;j<subcount[1];++j) {
          for(i=0;i<subcount[0];++i) {
            windex = fcenter_windex + k*fwxydim + j*fwxdim + i;
            Hfine += fwMs[windex] * (*fwoed.field)[windex];
          }
        }
      }
      Hfine *= 1.0/(cwMs[ccenter_windex]*subcount[0]*subcount[1]*subcount[2]);
      Hdiffz = Hfine - Hcoarse;
    }

#ifdef DEBUG_PRINT
    printf("Coarse index (%2d,%2d,%2d) ---\n",icoarse,jcoarse,kcoarse);
    printf(" Hdiff = %21.15g %21.15g %21.15g\n"
           "         %21.15g %21.15g %21.15g\n"
           "         %21.15g %21.15g %21.15g\n\n",
           Hdiffx.x,Hdiffy.x,Hdiffz.x,
           Hdiffx.y,Hdiffy.y,Hdiffz.y,
           Hdiffx.z,Hdiffy.z,Hdiffz.z);
    fflush(stdout);
#endif // DEBUG_PRINT

    // Use H data to compute compensatory demag interaction tensor
    REAL8m& Nxx = Ncorr_diag[cindex].x;
    REAL8m& Nyy = Ncorr_diag[cindex].y;
    REAL8m& Nzz = Ncorr_diag[cindex].z;
    REAL8m& Nxy = Ncorr_offdiag[cindex].x;
    REAL8m& Nxz = Ncorr_offdiag[cindex].y;
    REAL8m& Nyz = Ncorr_offdiag[cindex].z;
    Nxx = Hdiffx.x;
    Nyy = Hdiffy.y;
    Nzz = Hdiffz.z;

    // The tensor should be symmetric, e.g., Hdiffx.y should equal
    // Hdiffy.x, but use the average anyway.  The assert checks
    // below require relative error smaller than 1e-8, which sounds
    // like a reasonable value but has no theoretical or empirical
    // reasoning behind it.
    Nxy = 0.5*(Hdiffx.y+Hdiffy.x);
    assert(fabs(Hdiffx.y-Hdiffy.x) <= 1e-8*fabs(Nxy));
    Nxz = 0.5*(Hdiffx.z+Hdiffz.x);
    assert(fabs(Hdiffx.z-Hdiffz.x) <= 1e-8*fabs(Nxz));
    Nyz = 0.5*(Hdiffy.z+Hdiffz.y);
    assert(fabs(Hdiffy.z-Hdiffz.y) <= 1e-8*fabs(Nyz));
  }

}

// Constructor
Oxs_EdgeCorrAnisotropy::Oxs_EdgeCorrAnisotropy(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr)
{
  // Process arguments

  // Create full coarse mesh, and set read lock
  OXS_GET_INIT_EXT_OBJECT("mesh",Oxs_RectangularMesh,coarse_mesh);
  coarse_mesh_key.Set(coarse_mesh.GetPtr());  // Sets a dep lock
  const Oxs_RectangularMesh& cmesh = coarse_mesh_key.GetReadReference();


  // Program logic for setting Ncorr values:
  //   The Ncorr_diag and Ncorr_offdiag Oxs_MeshValue arrays
  //   can be set either from a file ("readN" init value) or
  //   by computation ("supercount" et al. init values).  If
  //   "readN" is given and "supercount" is not, then the Ncorrs
  //   will be filled from files, with an exception thrown on
  //   error.  If both "readN" and "supercount" are given, then
  //   Ncorrs will be filled from file if possible, but if the
  //   read fails for any reason then Ncorrs will be computed
  //   instead.  If "readN" is not given by "supercount" is,
  //   then Ncorrs will be computed.  It is a fatal error if
  //   neither "readN" nor "supercount" are specified in the
  //   MIF file.
  //
  //   The "saveN" init value is ignored if Ncorrs were filled
  //   from file.  It is not considered an error to specify
  //   "saveN" w/o also specifying "supercount", even though
  //   in that case "saveN" is superfluous.
  //
  //   Similarly, the "demag" init value is ignored if Ncorrs
  //   were filled from file.
  //
  //   In any case, Ms_coarse is always computed, which means
  //   that "Ms" is a required input, in all cases.

  // Input Ncorr values from a file?
  BOOL Ncorr_set_from_file = 0;
  String readN = GetStringInitValue("readN","");
  if(readN.length()>0) {
    // Oxs_MeshValue<ThreeVector> Ncorr_diag;
    // Oxs_MeshValue<ThreeVector> Ncorr_offdiag;
    String diagname    = readN;      diagname += String("-diag.ovf");
    String offdiagname = readN;   offdiagname += String("-offdiag.ovf");

    Vf_FileInput* vffi=NULL;
    Vf_Mesh* file_mesh=NULL;

    // Ncorr_diag
    Nb_DString dummy;
    vffi=Vf_FileInput::NewReader(diagname.c_str(),&dummy);
    if(vffi!=NULL) {
      file_mesh = vffi->NewMesh();
      delete vffi;
      if(file_mesh==NULL || !cmesh.IsCompatible(file_mesh)) {
        if(file_mesh!=NULL) delete file_mesh;
      } else {
        cmesh.FillMeshValueExact(file_mesh,Ncorr_diag);
        delete file_mesh;

        // Ncorr_offdiag
        vffi=Vf_FileInput::NewReader(offdiagname.c_str(),&dummy);
        if(vffi!=NULL) {
          file_mesh = vffi->NewMesh();
          delete vffi;
          if(file_mesh==NULL || !cmesh.IsCompatible(file_mesh)) {
            if(file_mesh!=NULL) delete file_mesh;
          } else {
            cmesh.FillMeshValueExact(file_mesh,Ncorr_offdiag);
            delete file_mesh;
            Ncorr_set_from_file = 1;
          }
        }
      }
    }
  }

  INT4m supercount[3];
  supercount[0] = supercount[1] = supercount[2] = 0;
  if(Ncorr_set_from_file) {
    // Ignore "supercount", if any
    DeleteInitValue("supercount");
  } else {
    ThreeVector vsup = GetThreeVectorInitValue("supercount");
    if(vsup.x != floor(vsup.x) || vsup.x<1.0 ||
       vsup.y != floor(vsup.y) || vsup.y<1.0 ||
       vsup.z != floor(vsup.z) || vsup.z<1.0) {
      throw Oxs_Ext::Error(this,"Inputs to supercount must be"
                           " positive integers.");
    }
    supercount[0] = static_cast<int>(floor(vsup.x));
    supercount[1] = static_cast<int>(floor(vsup.y));
    supercount[2] = static_cast<int>(floor(vsup.z));
  }


  ThreeVector vsub = GetThreeVectorInitValue("subcount");
  if(vsub.x != floor(vsub.x) || vsub.x<1.0 ||
     vsub.y != floor(vsub.y) || vsub.y<1.0 ||
     vsub.z != floor(vsub.z) || vsub.z<1.0) {
    throw Oxs_Ext::Error(this,"Inputs to subcount must be"
			 " positive integers.");
  }
  INT4m subcount[3];
  subcount[0] = static_cast<int>(floor(vsub.x));
  subcount[1] = static_cast<int>(floor(vsub.y));
  subcount[2] = static_cast<int>(floor(vsub.z));

  // Scalar field used for filling fine mesh
  Oxs_OwnedPointer<Oxs_ScalarField> Ms_fine_init;
  OXS_GET_INIT_EXT_OBJECT("Ms",Oxs_ScalarField,Ms_fine_init);

  if(Ncorr_set_from_file) {
    // Fill Ms_coarse
    ComputeMsCoarse(director,subcount,
                    coarse_mesh_key.GetPtr(),
                    Ms_fine_init.GetPtr(),
                    Ms_coarse);
    DeleteInitValue("demag");
  } else {
    // Fill Ms_coarse, Ncorr_diag, and Ncorr_offdiag

    // Create two demag objects, one for coarse working mesh
    // and one for fine working mesh.
    vector<String> demag_spec;
    Oxs_SplitList demag_pieces;
    String demag_init = GetStringInitValue("demag","Oxs_Demag {}");
    demag_pieces.Split(demag_init);
    demag_pieces.FillParams(demag_spec);
    Oxs_OwnedPointer<Oxs_Energy> cDemag;
    OXS_GET_EXT_OBJECT(demag_spec,Oxs_Energy,cDemag);
    Oxs_OwnedPointer<Oxs_Energy> fDemag;
    OXS_GET_EXT_OBJECT(demag_spec,Oxs_Energy,fDemag);

    ComputeAnisotropyCorrections(director,subcount,supercount,
                                 cDemag.GetPtr(),fDemag.GetPtr(),
                                 coarse_mesh_key.GetPtr(),
                                 Ms_fine_init.GetPtr(),
                                 Ms_coarse,Ncorr_diag,Ncorr_offdiag);
  }
                               
  String saveN = GetStringInitValue("saveN","");
  if(saveN.length()>0 && !Ncorr_set_from_file) {
    // Oxs_MeshValue<ThreeVector> Ncorr_diag;
    // Oxs_MeshValue<ThreeVector> Ncorr_offdiag;
    String diagname    = saveN;      diagname += String("-diag.ovf");
    String offdiagname = saveN;   offdiagname += String("-offdiag.ovf");
    coarse_mesh_key.GetPtr()->Oxs_Mesh::WriteOvf(diagname.c_str(),1,
       "N-diag",
       "Oxs_EdgeCorrAnisotropy adjustment anisotropy coefficients:"
       " Nxx, Nyy, Nzz",
       "A/m","rectangular","binary","8",&Ncorr_diag,NULL);
    coarse_mesh_key.GetPtr()->Oxs_Mesh::WriteOvf(offdiagname.c_str(),1,
       "N-offdiag",
       "Oxs_EdgeCorrAnisotropy adjustment anisotropy coefficients:"
       " Nxy, Nxz, Nyz",
       "A/m","rectangular","binary","8",&Ncorr_offdiag,NULL);
  }

  // Create Oxs_ScalarField object to hold Ms values on coarse mesh.
  // This is then passed off to the director to hold.
  String cMs_name = String("Oxs_ArrayScalarField:")
    + GetStringInitValue("coarsename",String("coarseMs"));
  Oxs_Box bbox;
  cmesh.GetBoundingBox(bbox);
  Oxs_ArrayScalarField* cMs
    = new Oxs_ArrayScalarField(cMs_name.c_str(),director,"");
  Oxs_OwnedPointer<Oxs_Ext> cMs_holder;
  cMs_holder.SetAsOwner(static_cast<Oxs_Ext*>(cMs)); // Handles clean-up
  const UINT4m imax_coarse = cmesh.DimX();
  const UINT4m jmax_coarse = cmesh.DimY();
  const UINT4m kmax_coarse = cmesh.DimZ();
  cMs->InitArray(imax_coarse,jmax_coarse,kmax_coarse,bbox);
  for(UINT4m k=0;k<kmax_coarse;++k) {
    for(UINT4m j=0;j<jmax_coarse;++j) {
      for(UINT4m i=0;i<imax_coarse;++i) {
        cMs->SetArrayElt(i,j,k,Ms_coarse[cmesh.Index(i,j,k)]);
      }
    }
  }
  director->StoreExtObject(cMs_holder); // Transfer ownership to director.
  /// This also puts cMs on the director's ExtObject list.


  VerifyAllInitArgsUsed();
}

BOOL Oxs_EdgeCorrAnisotropy::Init()
{
  return Oxs_Energy::Init();
}

void Oxs_EdgeCorrAnisotropy::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  UINT4m size = state.mesh->Size();
  if(coarse_mesh->Id() != state.mesh->Id()) {
    // State mesh must be the coarse mesh used during object construction.
    String msg = String("State mesh from driver is not the same"
                        " mesh used to initialize object ");
    msg += String(InstanceName());
    msg += String(".");
    throw Oxs_Ext::Error(this,msg);
  }

  for(UINT4m i=0;i<size;++i) {
    REAL8m Msat = Ms[i];
    if(Msat==0.0) {
      energy[i]=0.0;
      field[i].Set(0.,0.,0.);
      continue;
    }
    // N correction kludge. N has -1*Msat built-in.
    // See also notes in edgecorranisotropy.h.
    const REAL8m& Nxx = Ncorr_diag[i].x;
    const REAL8m& Nyy = Ncorr_diag[i].y;
    const REAL8m& Nzz = Ncorr_diag[i].z;
    const REAL8m& Nxy = Ncorr_offdiag[i].x;
    const REAL8m& Nxz = Ncorr_offdiag[i].y;
    const REAL8m& Nyz = Ncorr_offdiag[i].z;

    const ThreeVector& m = spin[i];

    field[i].x = (Nxx*m.x + Nxy*m.y + Nxz*m.z);
    field[i].y = (Nxy*m.x + Nyy*m.y + Nyz*m.z);
    field[i].z = (Nxz*m.x + Nyz*m.y + Nzz*m.z);

    energy[i] = -0.5*MU0*Msat*(field[i]*m);
  }
}
