/* FILE: lagrangeevolve.cc                 -*-Mode: c++-*-
 *
 * Concrete minimization evolver class, using conjugate gradients
 * applied to augmented Lagrangian, as suggested by Tony Kearsley,
 * 25-Oct-2001.
 */

#include <float.h>
#include <string>

#include "nb.h"
#include "director.h"
#include "mindriver.h"
#include "simstate.h"
#include "lagrangeevolve.h"
#include "key.h"
#include "threevector.h"
#include "energy.h"             // Needed to make MSVC++ 5 happy

OC_USE_STRING;

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_LagrangeEvolve);

/* End includes */

// Constructor
Oxs_LagrangeEvolve::Oxs_LagrangeEvolve(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_MinEvolver(name,newdtr,argstr),
    current_id(0),current_L(0.),
    working_stage_number(0),
    lambda_valid(0),
    lambda_CG_cycle_count(0),
    lambda_CG_cycle_max_L_drop(0.),
    base_valid(0),base_use_conjugate_direction(0),
    base_id(0),base_L(0.),
    base_direction_max_mag(1.),base_dL_normsq(0.),
    bracket_valid(0),
    last_minimum_reduction_ratio(0.5),
    next_to_last_minimum_reduction_ratio(0.5),
    energy_calc_count(0),
    lambda_cycle_count(0),direction_cycle_count(0),
    bracket_cycle_count(0),linemin_cycle_count(0)
{
  // Process arguments
  rho = GetRealInitValue("rho",1.0); // Weighting on Lagrangian
  /// augmentation term.
  normsq_guard = GetRealInitValue("normsq_guard",0.25); // Don't allow
  /// non-unit x[i]'s norm squared drifted farther from 1.0 than this.

  REAL8m min_ang = GetRealInitValue("minimum_bracket_step",0.05);
  min_ang *= PI/180.; // Convert from degrees to radians
  minimum_bracket_step = tan(min_ang);
  bracket_start_step = minimum_bracket_step; // Dummy init value

  REAL8m max_ang = GetRealInitValue("maximum_bracket_step",10);
  max_ang *= PI/180.; // Convert from degrees to radians
  maximum_bracket_step = tan(max_ang);

  line_minimum_relwidth
    = GetRealInitValue("line_minimum_relwidth",1e-6);

  lambda_min_CG_cycle_count
    = GetUIntInitValue("lambda_min_CG_cycle_count",5);

  lambda_reset_L_drop_ratio
    = GetRealInitValue("lambda_reset_L_drop_ratio",0.01);

  CG_collapse_guard
    = GetRealInitValue("CG_collapse_guard",0.02);

  // Setup outputs
  max_mxHxm_output.Setup(this,InstanceName(),"Max mxHxm","A/m",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);
  delta_E_output.Setup(this,InstanceName(),"Delta E","J/s",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);
  total_energy_output.Setup(this,InstanceName(),"Total energy","J",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);
  L_output.Setup(this,InstanceName(),"Augmented Lagrangian","",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);
  lambda_cycle_count_output.Setup(this,InstanceName(),
     "Lambda cycle count","",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);
  direction_cycle_count_output.Setup(this,InstanceName(),
     "Direction cycle count","",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);
  bracket_cycle_count_output.Setup(this,InstanceName(),
     "Bracket cycle count","",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);
  linemin_cycle_count_output.Setup(this,InstanceName(),
     "Linemin cycle count","",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);
  energy_calc_count_output.Setup(this,InstanceName(),
     "Energy calc count","",0,
     &Oxs_LagrangeEvolve::UpdateDerivedOutputs);

  // Register outputs
  max_mxHxm_output.Register(director,-5);
  delta_E_output.Register(director,-5);
  total_energy_output.Register(director,-5);
  L_output.Register(director,-5);
  lambda_cycle_count_output.Register(director,-5);
  direction_cycle_count_output.Register(director,-5);
  bracket_cycle_count_output.Register(director,-5);
  linemin_cycle_count_output.Register(director,-5);
  energy_calc_count_output.Register(director,-5);

  VerifyAllInitArgsUsed();
}

BOOL Oxs_LagrangeEvolve::StageInit()
{
  weighted_rho=0.; // Dummy init value
  lambda_valid=0;
  lambda_CG_cycle_count=0;
  lambda_CG_cycle_max_L_drop=0.;
  current_id=0;
  current_L=0.;  // Dummy init value
  base_valid=0;
  base_use_conjugate_direction=0;
  base_id=0;
  base_L=0.;
  base_direction_max_mag=1.0;   // Dummy init value
  base_dL_normsq=0.0;
  bracket_valid=0;
  last_minimum_reduction_ratio=0.5;
  next_to_last_minimum_reduction_ratio=0.5;
  bracket_start_step = minimum_bracket_step;
  return 1;
}

BOOL Oxs_LagrangeEvolve::Init()
{
  if(!Oxs_MinEvolver::Init()) return 0;

  // Initialize instance variables  
  working_stage_number=0;
  energy_calc_count=0;
  lambda_cycle_count=0;
  direction_cycle_count=0;
  bracket_cycle_count=0;
  linemin_cycle_count=0;

  return StageInit();
}

Oxs_LagrangeEvolve::~Oxs_LagrangeEvolve()
{}

REAL8m Oxs_LagrangeEvolve::GetEnergyAndmxHxm(
 const Oxs_SimState& state,          // Import
 Oxs_MeshValue<REAL8m>& energy,      // Export
 Oxs_MeshValue<ThreeVector>& mxHxm)  // Export
{ // Note: (mxH)xm = mx(Hxm)
  UINT4m i;

  // For convenience
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  const Oxs_Mesh* mesh = state.mesh;
  UINT4m vecsize = mesh->Size();

  // Export space
  energy.AdjustSize(mesh);
  mxHxm.AdjustSize(mesh);

  ++energy_calc_count; // Update count

  REAL8m max_torque_sq=0.0;
  UINT4m energy_obj_count = director->GetEnergyObjCount();
  UINT4m energy_obj_index = 0;
  if(energy_obj_index<energy_obj_count) {
    {
      Oxs_EnergyData oed(state);
      oed.energy_buffer = &energy;
      oed.field_buffer  = &mxHxm;
      director->GetEnergyObj(energy_obj_index)->GetEnergyData(state,oed);
      if(oed.pE_pt.IsSet() && oed.pE_pt != 0.0) {
        String msg = 
          String("Oxs_LagrangeEvolve::GetEnergyAndmxHxm: Energy object ")
          + director->GetEnergyObj(energy_obj_index)->InstanceName()
          + String(" is time varying; this property is not supported "
                   " my minimization evolvers.");
        throw Oxs_Ext::Error(this,msg.c_str());
      }
      if(oed.energy.Get() != &energy) {
        energy = *oed.energy; // oed.energy_buffer wasn't used, so copy
        /// from actual buffer space to energy.
      }
      if(oed.field.Get() != &mxHxm) {
        for(i=0;i<vecsize;++i) {
          ThreeVector temp = (*oed.field)[i];
          temp ^= spin[i];
          mxHxm[i] = temp;  // Right now this holds Hxm
        }
      } else {
        for(i=0;i<vecsize;++i) {
          mxHxm[i] ^= spin[i];  // Right now this holds Hxm
        }
      }
    }
    while(++energy_obj_index<energy_obj_count) {
      Oxs_EnergyData oed(state);
      oed.energy_buffer = &scratch_energy;
      oed.field_buffer  = &scratch_field;
      if(oed.energy_buffer == &energy ||
	 oed.field_buffer == &mxHxm) {
	String msg = "Oxs_LagrangeEvolve::GetEnergyAndmxHxm:"
	  " scratch buffer overwrite detected.";
	throw Oxs_Ext::Error(this,msg.c_str());
      }
      director->GetEnergyObj(energy_obj_index)->GetEnergyData(state,oed);
      energy += *oed.energy;
      for(i=0;i<vecsize;++i) {
        ThreeVector temp = (*oed.field)[i];
        temp ^= spin[i];
        mxHxm[i] += temp;  // Accumulating Hxm
      }
    }

    // Zero torque at fixed spin sites.
    UpdateFixedSpinList(mesh);
    const UINT4m fixed_count = GetFixedSpinCount();
    for(UINT4m j=0;j<fixed_count;j++) {
      mxHxm[GetFixedSpin(j)].Set(0.,0.,0.);
    }

    // Replace Hxm with mxHxm, and calculate max_torque
    max_torque_sq=0.0;
    for(i=0;i<vecsize;++i) {
      if(Ms[i]==0.0) { // Ignore zero-moment spins
        mxHxm[i].Set(0.,0.,0.);
        energy[i] = 0.;
      } else {
	ThreeVector Hxm = mxHxm[i];  // Currently Hxm
        REAL8m torque_sq = Hxm.MagSq();
        if(torque_sq>max_torque_sq) max_torque_sq = torque_sq;
        ThreeVector temp = spin[i];
        temp ^= Hxm;
        mxHxm[i] = temp;
      }
    }
  } else {
    // No energy terms!
    energy = 0;
    mxHxm = ThreeVector(0.,0.,0.);
    max_torque_sq=0.0; // Safety
  }
  state.AddDerivedData("Max mxHxm",sqrt(max_torque_sq));
  // Note: |mxH| = |mxHxm| provided |m|=1.

  Nb_Xpfloat total_energy = 0.0;
  for(i=0;i<vecsize;++i) {
    total_energy.Accum(energy[i] * mesh->Volume(i));
  }
  state.AddDerivedData("Total energy",total_energy.GetValue());

  // Append supplemental derived data
  state.AddDerivedData("Augmented Lagrangian",current_L);
  state.AddDerivedData("Energy calc count",
		       REAL8m(energy_calc_count));
  state.AddDerivedData("Lambda cycle count",
		       REAL8m(lambda_cycle_count));
  state.AddDerivedData("Direction cycle count",
		       REAL8m(direction_cycle_count));
  state.AddDerivedData("Bracket cycle count",
		       REAL8m(bracket_cycle_count));
  state.AddDerivedData("Linemin cycle count",
		       REAL8m(linemin_cycle_count));

  return total_energy;
}

void Oxs_LagrangeEvolve::EvaluateLagrangian
(const Oxs_SimState&  state,
 const Oxs_MeshValue<ThreeVector>& x,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& mxHxm,
 REAL8m& rel_L, REAL8m& Lp)
{ // Note: See also Oxs_LagrangeEvolve::Calculate_dL()
  //       and Oxs_LagrangeEvolve::UpdateCurrentData()
  GetEnergyAndmxHxm(state,energy,mxHxm);

  // L = \sum_i (E_i + lambda_i*q_i + 0.5*weighted_rho*q_i^2)*V_i,
  // where q is vector of residuals, q[i] = |x[i]|^2 - 1,
  // E is energy density and V is cell volume.
  //
  // rel_L = L(current) - L(basept)
  // dL[i]=(-mu0*Ms[i]H[i] + 2 (lambda[i] + weighted_rho*q[i])x[i] )V[i]
  //    where we use mxHxm/|x| in place of H since we feed
  //    m (as opposed to non-normalized x) into E, hence
  //    there is in effect no change in E in the m direction,
  //    so effective H should be orthogonal to m.  See mjd's
  //    NOTES II, 4-June-2002 p 142 (Errata) and 29-May-2002 p 156.
  //
  // Lp = dL . base_direction
  //
  const Oxs_Mesh* mesh = state.mesh;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  const UINT4m vecsize = mesh->Size();
  rel_L=Lp=0;
  Nb_Xpfloat rel_L_xp=0.,Lp_xp=0.;
  for(UINT4m i=0;i<vecsize;++i) {
    // Calculate contribution to L from cell i.
    REAL8m xnormsq = x[i].MagSq();
    REAL8m xscale = 1.0;
    if(xnormsq>0) xscale = 1.0/sqrt(xnormsq);
    REAL8m base_xnormsq = base_x[i].MagSq();
    REAL8m normsq_diff = xnormsq-base_xnormsq;
    REAL8m tempsum = lambda[i]
      + 0.5*weighted_rho*((xnormsq-1.0)+(base_xnormsq-1.0));
    tempsum *= normsq_diff;
    tempsum += (energy[i] - base_energy[i]);
    const REAL8m volume_i = mesh->Volume(i);
    rel_L_xp += (tempsum * volume_i);

    // Calculate contribution to Lp from cell i.
    tempsum = 2*(lambda[i]
		 + weighted_rho*(xnormsq-1))*(x[i]*base_direction[i]);
    tempsum += (-MU0 * Ms[i] * xscale) * (mxHxm[i]*base_direction[i]);
    /// Should this be * Ms[i] ??? /**/
    Lp_xp += tempsum*volume_i;
  }
  rel_L = rel_L_xp;
  Lp = Lp_xp;
}

void Oxs_LagrangeEvolve::FillBracket
(REAL8m offset,                             // Imports...
 const Oxs_SimState* oldstate,
 Oxs_Key<Oxs_SimState>& new_state_key,      // Exports...
 Bracket& endpt)
{ // Also fills cache space with new_state data
  try {
    Oxs_SimState& workstate
      = new_state_key.GetWriteReference(); // Write lock
    Oxs_MeshValue<ThreeVector>& spin = workstate.spin;
    UINT4m vecsize = workstate.mesh->Size();
    cache_id = 0;
    cache_x.AdjustSize(workstate.mesh);
    REAL8m max_non_unit=0.0;
    for(UINT4m i=0;i<vecsize;i++) {
      ThreeVector temp = base_x[i] + offset * base_direction[i];
      cache_x[i] = temp;
      REAL8m non_unit = fabs(temp.MagSq() - 1.0);
      if(non_unit>max_non_unit) max_non_unit = non_unit;
      temp.MakeUnit();
      spin[i] = temp;
    }
    workstate.iteration_count = oldstate->iteration_count + 1;
    workstate.stage_iteration_count
      = oldstate->stage_iteration_count + 1;

    const Oxs_SimState& newstate
      = new_state_key.GetDepReference(); // Release write lock
    newstate.AddDerivedData("Max norm error",max_non_unit);
    cache_id = newstate.Id();
    EvaluateLagrangian(newstate,cache_x,
                       cache_energy,cache_mxHxm,
                       endpt.rel_L,endpt.Lp);
    cache_L = base_L + endpt.rel_L;
    endpt.offset=offset;
  }
  catch (...) {
    new_state_key.GetReadReference(); // Release write lock
    throw;
  }
}

BOOL Oxs_LagrangeEvolve::FindBracketStep
(const Oxs_SimState* oldstate,
 Oxs_Key<Oxs_SimState>& statekey,
 BOOL& bracket_error)
{ // Takes one step; assumes
  //     leftpt.offset <= midpt.offset <= rightpt.offset
  // and minimum to be bracketed is to the right of rightpt.offset.
  // Returns TRUE if minimum is successfully bracketed, otherwise
  // FALSE.  Also sets bracket_error to TRUE on failure to bracket
  // minima inside allowed bracket range.
  bracket_error=0;

  REAL8m step_scaling = 1.0;
  if(base_direction_max_mag>0.0) {
    step_scaling = 1.0/base_direction_max_mag;
  }

  REAL8m offset = rightpt.offset;
  if(offset < step_scaling*bracket_start_step) {
    offset = step_scaling*bracket_start_step;
  } else {
    if(offset<midpt.offset+step_scaling*minimum_bracket_step) {
      offset = midpt.offset*2;
    } else {
      // Try to use derivate information at existing endpoints
      // to estimate location of minimum.
      REAL8m temp1 = (midpt.offset-rightpt.offset)*midpt.Lp;
      REAL8m temp2 = rightpt.Lp-midpt.Lp;
      if(temp2<=0) {
        offset *= 4; // Aiee! We're rolling off a cliff!
      } else {
        REAL8m temp3 = 8*rightpt.offset-midpt.offset;
        if(2*temp1 < temp3*temp2) { // Overflow check
          offset = midpt.offset + 2*temp1/temp2;
        } else {
          offset = 8*rightpt.offset;
        }
      }
    }
  }
  if(offset < 2*rightpt.offset) {
    offset = 2*rightpt.offset;
  }
  if(offset > step_scaling * maximum_bracket_step) {
    offset = step_scaling * maximum_bracket_step;
  }

  // Zero span check
  if(offset <= rightpt.offset) {
    // Proposed bracket span increase is floating point 0.
    if(rightpt.offset>0) {
      offset = rightpt.offset*(1+16*REAL8_EPSILON);
    } else {
      offset = step_scaling * maximum_bracket_step;
      if(offset<=0) {
	// Punt
	offset = REAL8_EPSILON;
      }
    }
  }
  if(offset*base_direction_max_mag<16*REAL8_EPSILON) {
    // If offset*base_direction<REAL8_EPSILON, and we assume
    // that the base spins are unit vectors, then there is a
    // good chance that the proposed offset won't actually
    // change any of the discretized spins.  For bracketing
    // purposes, it shouldn't hurt to bump this up a bit.
    if(base_direction_max_mag >= 0.5 ||
       DBL_MAX*base_direction_max_mag > 16*REAL8_EPSILON) {
      offset = 16*REAL8_EPSILON/base_direction_max_mag;
    }
  }

  Bracket temppt;
  FillBracket(offset,oldstate,statekey,temppt);

  // Shift points to the right.  One might want to add some
  // check here to see that L is decreasing as predicted;
  // if not, then perhaps we've overstepped the minimum.
  // (Note: Depending on the size of Bracket, it might
  // be cheaper to drop the checks and just do the copies,
  // e.g., leftpt=midpt, in all cases.)
  if(leftpt.offset<midpt.offset)  leftpt=midpt;
  if(midpt.offset<rightpt.offset) midpt=rightpt;
  rightpt=temppt;

  // Classify situation.  We always require leftpt and midpt have Lp<0,
  // and leftpt.rel_L>=midpt.rel_L, so there are two possibilities:
  //  1) rightpt.rel_L>midpt.rel_L or rightpt.Lp>=0 ==> Minimum is bracketed
  //  2) rightpt.rel_L<midpt.rel_L and rightpt.Lp<0 ==> Minimum not bracketed
  BOOL bracket_success=0;
  if(rightpt.rel_L<midpt.rel_L && rightpt.Lp<0) {
    // Minimum not bracketed.
    if(rightpt.offset >= step_scaling * maximum_bracket_step) {
      // Unable to bracket minimum inside allowed range.
      // Shift leftpt and midpt to rightpt to disable line minimizer.
      bracket_error=1;
      leftpt=midpt=rightpt;
    }
  } else {
    // Minimum is bracketed.
    bracket_success=1;
  }

  return bracket_success;
}

BOOL Oxs_LagrangeEvolve::TakeLineMinimumStep
(const Oxs_SimState* old_state,     // Import
 Oxs_Key<Oxs_SimState>& state_key)
{ // Compresses bracket span. Assumes coming in that minimum lies
  // between leftpt.offset and rightpt.offset, and that lefpt.Lp<0.
  // These restrictions are maintained on exit.  Export state_key
  // is filled with next test state, and cache structures are filled.
  //    If this state has lower L than both previous brackets, then
  // return value is TRUE, otherwise false.

  BOOL new_min_L = 0;
  REAL8m rightspan = rightpt.offset - midpt.offset;
  REAL8m leftspan = midpt.offset - leftpt.offset;

  REAL8m alpha = 0.5; // Default (bisection).  Measured from midpt.
  BOOL right_minimum = 1; // If true, minimum is in righthand interval
  if(leftspan>0. && (midpt.Lp>=0. || midpt.rel_L>=leftpt.rel_L)) {
    right_minimum=0;
  }
  // NB: right_minimum is true iff test_offset>=0.

  REAL8m smallspan = fabs(midpt.offset);
  if(smallspan<=0) smallspan = fabs(rightpt.offset);
  smallspan *= 256*REAL8_EPSILON;
  // Any span smaller than smallspan is numerically suspect

  // Sort by case
  if(midpt.Lp==0) {
    alpha=0.;
  } else if(leftspan<=smallspan || rightspan<=smallspan ||
	    (!right_minimum && rightspan>6*leftspan) ||
	    (right_minimum && leftspan>6*rightspan)) {
    // Either one span is so small that roundoff errors can
    // lead to very bad parabolic fits (see next arm), or else
    // auxiliary point in away interval is far removed from
    // minimum.  So work with data surrounding interval only.
    REAL8m L1,L1p,L2,L2p,span;
    if (!right_minimum) {
      // Minimum in left span.  As usual, leftpt.Lp<0.
      L1 = leftpt.rel_L;   L1p = leftpt.Lp;
      L2 = midpt.rel_L;    L2p = midpt.Lp;
      span = leftspan;
    } else {
      // Minimum in right span.  In this case, midpt.Lp<0
      L1 = midpt.rel_L;    L1p = midpt.Lp;
      L2 = rightpt.rel_L;  L2p = rightpt.Lp;
      span = rightspan;
    }
    L1p *= span;
    L2p *= span;
    REAL8m Ldiff = L2 - L1;
    BOOL poor_data = 0;
    if(fabs(Ldiff)<=1e-10*(fabs(L1)+fabs(L2))) {
      poor_data = 1; // Data is likely unreliable
    }
    if(!poor_data) {
      // Cubic fit.  See mjd's NOTES II, 1-Sep-2001, p134-136.
      REAL8m a = -2*Ldiff +   L1p + L2p;
      REAL8m b =  3*Ldiff - 2*L1p - L2p;
      REAL8m c = L1p;
      if(a==0.0) {
	// Quadratic
	if(b!=0.0) { // Safety check. b should be >= -c/2 > 0.
	  alpha = -c/(2*b);
	}
      } else {
	REAL8m disc = b*b - 3*a*c;
	if(disc>=0.0) { // Safety check.  See NOTES II, 1-Sep-2001, p135.
	  disc = sqrt(disc);
	  if(b>=0.) {
	    alpha = -c/(b + disc);
	  } else {
	    alpha = (-b + disc)/(3*a);
	  }
	}
      }
    }
    // Check aggressive reductions and cases with suspect
    // data against lower order method.
    const double AGGRESSIVE_FIT = 0.0625;
    if(poor_data ||
       alpha<AGGRESSIVE_FIT || alpha>1-AGGRESSIVE_FIT) {
      // Calculate alpha using lower-order method
      REAL8m alt_alpha = 0.5;
      if(L2p<0.) {
	// Overly large bracket, or bad right Lp.
	// Guess at minpt using L1/L1p and L2
	const REAL8m reduce_limit = AGGRESSIVE_FIT*0.25; // The
	/// situation rightpt.Ep<0 is a bit suspicious, so limit
	/// reduction.
	REAL8m numerator = -1.0 * L1p; // This is > 0.
	REAL8m denominator = 2*(Ldiff - L1p);
	// denominator must also be >0, because L2>=L1
	// if a minimum is bracketed with L2p<0.
	denominator = fabs(denominator); // But play it safe, because
	/// checks below depend on denominator>0.
	if(numerator<reduce_limit*denominator) {
	  alt_alpha = reduce_limit;
	} else if(numerator>(1-reduce_limit)*denominator) {
	  alt_alpha = 1-reduce_limit;
	} else {
	  alt_alpha = numerator/denominator;
	}
      } else {
	// Use just derivative information, i.e., quadratic fit.
	REAL8m Lp_diff = L2p - L1p; // Since L1p<0 and L2p>=0,
	/// Lp_diff = |L2p| + |L1pt.Ep|.
	alt_alpha = -L1p/Lp_diff;
	// Provided Lp_diff is a valid floating point number,
	// then alt_alpha should not overflow, and in fact must
	// lay inside [0,1].
      } // matches if(L2p<0.)

      // Be conservative
      if(poor_data) {
	alpha = alt_alpha;
      } else if(alpha<=0.5) {
	if (alpha<alt_alpha) alpha = alt_alpha;
      } else {
	if (alpha>alt_alpha) alpha = alt_alpha;
      }
    } // Matches if(alpha<AGGRESSIVE_FIT || alpha>1-AGGRESSIVE_FIT)
    if (!right_minimum) {
      // alpha is measured with respect to midpt
      alpha = 1.0 - alpha;
    }
  } else {
    // Two distict intervals.  Fit parabola to the Lp's.
    // NOTES II, 1-Nov-2001, pp 143-146, except here we expand
    // about midpt rather than leftpt.  (NB: The interpolated
    // points are thus
    //                  leftpt-midpt = -1*leftspan
    //                   midpt-midpt = 0
    //                 rightpt-midpt = rightspan
    //
    REAL8m dp1 = leftpt.Lp - midpt.Lp;
    REAL8m dp2 = rightpt.Lp - midpt.Lp;
    REAL8m c = midpt.Lp;
    REAL8m denom = leftspan*rightspan*(leftspan+rightspan);
    REAL8m b = (leftspan*leftspan*dp2 - rightspan*rightspan*dp1)/denom;
    REAL8m a = (rightspan*dp1 + leftspan*dp2)/denom;

    BOOL root_error=1;
    // Find roots
    REAL8m disc = b*b - 4*a*c;
    REAL8m root1=0.,root2=0.;
    if(disc>=0.) {
      root_error=0;
      if(a==0.) { // Linear case
	if(fabs(b)>0.5 || fabs(c)<DBL_MAX*fabs(b)) { // Safety check
	  root1 = root2 = -c/b;
	} else {
	  root_error=1;
	}
      } else if(b>=0) {
	REAL8m temp = b+sqrt(disc);
	root1 = -0.5*temp/a;
	root2 = -2*c/temp;
      } else { // b<0
	REAL8m temp = b-sqrt(disc);
	root1 = -0.5*temp/a;
	root2 = -2*c/temp;
      }
    }
    if(!root_error) {
      if(root1>root2) {
	REAL8m temp=root1; root1=root2; root2=temp; // Order roots
      }
      if(!right_minimum) {
        // Looking for first root to the right of -leftspan
	REAL8m root=0;
	if(root1 > -leftspan) root=root1;
	else                  root=root2;
	alpha = root / (-leftspan);
      } else {
	// Looking for first root to the right of 0
	REAL8m root=0;
	if(root1 > 0) root=root1;
	else          root=root2;
	alpha = root / rightspan;
      }
      if(alpha<=0 || alpha>=1) {
	root_error=1;
	alpha=0.5;
      }
    }
    if(root_error) {
      // Above procedure did not yield a compatible solution,
      // so just use data at endpoints surrounding minimum.
      const REAL8m reduce_limit = 0.0625; // This case is
      // suspect, so apply hard limit to reduction.
      REAL8m L_a=0,Lp_a=0,L_b=0,Lp_b=0;
      if(!right_minimum) {
        L_a=leftpt.rel_L;
        Lp_a=leftpt.Lp;
        L_b=midpt.rel_L;
        Lp_b=midpt.Lp;
      } else {
        L_a=midpt.rel_L;
        Lp_a=midpt.Lp;
        L_b=rightpt.rel_L;
        Lp_b=rightpt.Lp;
      }
      alpha=0.5; // Relative distance for offset from pt a.
      if(Lp_a==0.) alpha=0.0;
      else if(Lp_a<0.) {
        if(Lp_b>=0) {
          // Do linear fit on derivative
          alpha = -Lp_a/(Lp_b - Lp_a);
        } else {
          // Fit quadratic through L_a, Lp_a and L_b
          REAL8m qnum = -Lp_a;
          REAL8m qdenom = L_b - L_a - Lp_a;
          if(qnum<qdenom) alpha = qnum/qdenom; // Note: qnum>0.
        }
      }
      if(alpha<reduce_limit)        alpha = reduce_limit;
      else if(alpha>1-reduce_limit) alpha = 1-reduce_limit;
      if(!right_minimum) {
        alpha = 1-alpha; // Alpha should be wrt midpt
      }
    }
  }

  // Limit reduction
  REAL8m reduce_low  = next_to_last_minimum_reduction_ratio;
  REAL8m reduce_high = last_minimum_reduction_ratio;
  if(reduce_low>reduce_high) {
    reduce_low  = last_minimum_reduction_ratio;
    reduce_high = next_to_last_minimum_reduction_ratio;
  }
  
  const REAL8m reduction_guard = 0.3;
  const REAL8m strong_reduction = 1-reduction_guard;
  const REAL8m poor_reduction = reduction_guard;
  const REAL8m overshoot_guard = 0.0625;
  if(reduce_high<strong_reduction) {
    if(reduce_low<poor_reduction) {
      // Neither of the last two steps produced a strong
      // reduction, and at least one was poor.
      if(alpha < reduction_guard)     alpha = reduction_guard;
      if(alpha > 1 - reduction_guard) alpha = 1 - reduction_guard;
    } else {
      // Neither of last two steps produced strong reductions,
      // but neither were poor.
      const REAL8m recovery_guard = 0.0625;
      if(alpha < recovery_guard)     alpha = recovery_guard;
      if(alpha > 1 - recovery_guard) alpha = 1 - recovery_guard;
    }
  } else if(last_minimum_reduction_ratio<overshoot_guard) {
    // Last try probably overstepped, so try cutting back on this pass.
    REAL8m adj = overshoot_guard*overshoot_guard;
    if(alpha<overshoot_guard)   alpha *= 1+adj;
    if(alpha>1-overshoot_guard) alpha *= 1-adj;
  }

  REAL8m test_offset=0.;
  if(!right_minimum) {
    if(alpha<=0.5) {
      REAL8m bump = alpha*leftspan;
      test_offset = midpt.offset - bump;
      if(test_offset >= midpt.offset) { // Out of bits
	test_offset = midpt.offset - REAL8_EPSILON*fabs(midpt.offset);
      }
    } else {
      REAL8m bump = (1-alpha)*leftspan;
      test_offset = leftpt.offset + bump;
      if(test_offset <= leftpt.offset) { // Out of bits
	test_offset = leftpt.offset + REAL8_EPSILON*fabs(leftpt.offset);
      }
    }
    if(test_offset<=leftpt.offset || test_offset>=midpt.offset) {
      // Final safety check
      alpha = 0.5;
      test_offset = 0.5 * (leftpt.offset + midpt.offset);
    }
  } else {
    if(alpha<=0.5) {
      REAL8m bump = alpha*rightspan;
      test_offset = midpt.offset + bump;
      if(test_offset <= midpt.offset) { // Out of bits
	test_offset = midpt.offset + REAL8_EPSILON*fabs(midpt.offset);
      }
    } else {
      REAL8m bump = (1-alpha)*rightspan;
      test_offset = rightpt.offset - bump;
      if(test_offset >= rightpt.offset) { // Out of bits
	test_offset = rightpt.offset - REAL8_EPSILON*fabs(rightpt.offset);
      }
    }
    if(test_offset<=midpt.offset || test_offset>=rightpt.offset) {
      // Final safety check
      alpha = 0.5;
      test_offset = 0.5 * (midpt.offset + rightpt.offset);
    }
  }

  Bracket testpt;
  FillBracket(test_offset,old_state,state_key,testpt);

  if(testpt.rel_L<leftpt.rel_L
     && testpt.rel_L<midpt.rel_L
     && testpt.rel_L<=rightpt.rel_L) {
    new_min_L=1;
  }

  // Update reduction ratio history only if a non-trivial
  // reduction was attempted.
  next_to_last_minimum_reduction_ratio = last_minimum_reduction_ratio;

  // Determine which intervals to keep.
  if(!right_minimum) {
    if(testpt.Lp>=0 || testpt.rel_L>=leftpt.rel_L) {
      // leftpt + testpt bracket a minimum, so keep the
      // leftmost intervals.  Note: It could happen that
      // other intervals would also pass this test, but
      // we give preference to the lefthand set because
      // they are closer to the basept.
      rightpt=midpt;
      midpt=testpt;
      last_minimum_reduction_ratio=alpha;
    } else {
      // Otherwise, minimum is between testpt and midpt,
      // and testpt.Lp<0, so we can keep either
      // [leftpt,testpt,midpt] or [testpt,midpt,rightpt]
      REAL8m testspan = testpt.offset-leftpt.offset;
      if((rightspan <= smallspan && testspan > smallspan) ||
	 (testspan <= rightspan
	  && (rightspan <= smallspan || testspan > smallspan))) {
        // Keep leftmost set
        rightpt=midpt;
        midpt=testpt;
      } else {
        // Keep rightmost set
        leftpt=testpt;
      }
      last_minimum_reduction_ratio=1-alpha;
    }
  } else {
    // If right_minimum, then midpt.Lp<0, so we can freely move
    // leftpt to midpt if desired.
    if(testpt.Lp<0 && testpt.rel_L<midpt.rel_L) {
      // Minimum is in rightmost interval, so pitch leftmost
      leftpt=midpt;
      midpt=testpt;
      last_minimum_reduction_ratio=alpha;
    } else {
      // Otherwise, minimum is between midpt and testpt, and
      // midpt.Lp<0, so we can keep either
      // [leftpt,midpt,testpt] or [midpt,testpt,rightpt]
      REAL8m testspan = rightpt.offset-testpt.offset;
      if((testspan <= smallspan && leftspan > smallspan) ||
	 (leftspan <= testspan
	  && (testspan <= smallspan || leftspan > smallspan))) {
        // Keep leftmost set
        rightpt=testpt;
      } else {
        // Keep rightmost set
        leftpt=midpt;
        midpt=testpt;
      }
      last_minimum_reduction_ratio=1-alpha;
    }
  }

  return new_min_L;
}

void Oxs_LagrangeEvolve::UpdateCurrentData(const Oxs_SimState& state)
{ // Note: See also Oxs_LagrangeEvolve::EvaluateLagrangian()
  current_id = 0; // Safety
  REAL8m E = GetEnergyAndmxHxm(state,current_energy,current_mxHxm);
  current_x = state.spin;  // Initialize with unit spins
  // L = (W + lambda.q + 0.5*weighted_rho*q.q).V, where q is vector
  //    of residuals, q[i] = |x[i]|^2 - 1, W is cell energy
  //    density, and V is cell volume.  See mjd's NOTES II,
  //    4-June-2002 p 142 (Errata) and 29-May-2002 p 156.
  // Since current_x holds unit spins, q[i]=0 for all i,
  // thus L = E.
  current_L = E;
  current_id = state.Id();
}

void
Oxs_LagrangeEvolve::FillLambdaFromCurrentData
(const Oxs_SimState& state)
{ // Sets up lambda from data in current data block.  Also
  // adjusts current_L to reflect the change in lambda, and
  // marks cache and base data invalid, since cache_L and
  // base_L would need to be updated, but aren't here.
  // Also resets x's to be unit vectors (?!),
  // and sets weighted_rho.
  const Oxs_Mesh* mesh = state.mesh;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  lambda.AdjustSize(mesh);
  const UINT4m vecsize = mesh->Size();
  Nb_Xpfloat new_L_xp=0;
  REAL8m max_wgtsq=0.0;
  for(UINT4m i=0;i<vecsize;i++) {
    current_x[i].MakeUnit(); // Do we really want to do this?
    REAL8m x_i_magsq = current_x[i].MagSq();
    REAL8m wgtsq = current_mxHxm[i].MagSq()*Ms[i]*Ms[i];
    if(wgtsq>max_wgtsq) max_wgtsq=wgtsq;
    lambda[i] = (0.5*MU0*Ms[i])*(current_x[i]*current_mxHxm[i])
      /(x_i_magsq*sqrt(x_i_magsq));
    REAL8m q_i = x_i_magsq - 1.0;
    REAL8m E_i = current_energy[i]
      + (lambda[i]+0.5*weighted_rho*q_i)*q_i;
    new_L_xp.Accum(E_i * mesh->Volume(i));
  }
  current_L = new_L_xp;
  weighted_rho = rho * MU0 * sqrt(max_wgtsq);
  cache_id = base_id = 0; // New lambda invalidates cache and base
  base_valid = 0;         // data.
}

REAL8m Oxs_LagrangeEvolve::Calculate_dL
(const Oxs_Mesh* mesh,                     // Imports...
 const Oxs_MeshValue<REAL8m>& Ms,
 const REAL8m work_rho,
 const Oxs_MeshValue<ThreeVector>& mxHxm,
 const Oxs_MeshValue<REAL8m> &work_lambda,
 const Oxs_MeshValue<ThreeVector>& x,
 Oxs_MeshValue<ThreeVector>& dL)           // Export
{ // On return, dL is filled, and the return value is
  // \sum_i |dL[i]|^2.  This routine does not use any
  // Oxs_LagrangeEvolve member variables.
  // dL[i] = (-mu0*Ms[i]*H[i] + 2 (lambda[i] + rho*q[i]) x[i])*V[i]
  //    where we use mxHxm/|x| in place of H since we feed
  //    m (as opposed to non-normalized x) into E, hence
  //    there is in effect no change in E in the m direction,
  //    so effective H should be orthogonal to m.
  // Note: See also Oxs_LagrangeEvolve::EvaluateLagrangian(),
  //  and mjd's NOTES II, 4-June-2002 p 142 (Errata) and
  //  29-May-2002 p 156.
  dL.AdjustSize(mesh);
  Nb_Xpfloat normsq_xp=0.;
  UINT4m vecsize = mesh->Size();
  for(UINT4m i=0;i<vecsize;i++) {
    dL[i] = x[i];
    REAL8m xnormsq = x[i].MagSq();
    REAL8m xscale = 1.0;
    if(xnormsq>0) xscale = 1.0/sqrt(xnormsq);
    dL[i] *= 2*(work_lambda[i]+work_rho*(xnormsq-1.));
    dL[i] += (-MU0*Ms[i]*xscale)*mxHxm[i];
    dL[i] *= mesh->Volume(i);
    normsq_xp.Accum(dL[i].MagSq());
  }
  return normsq_xp.GetValue();
}

REAL8m Oxs_LagrangeEvolve::FindMaxMag
(const Oxs_MeshValue<ThreeVector>& vec)
{
  UINT4m vecsize = vec.Size();
  REAL8m maxmagsq = 0.0;
  for(UINT4m i=0;i<vecsize;i++) {
    REAL8m magsq = vec[i].MagSq();
    if(magsq>maxmagsq) maxmagsq = magsq;
  }
  return sqrt(maxmagsq);
}

REAL8m Oxs_LagrangeEvolve::VecDot
(const Oxs_MeshValue<ThreeVector>& v,
 const Oxs_MeshValue<ThreeVector>& w)
{ // Support routine; returns \sum_i v_i*w_i
  UINT4m vecsize = v.Size();
  if(w.Size() != vecsize) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Oxs_LagrangeEvolve::VecDot:"
		" Vector size mismatch; v.Size=%u, w.Size=%u",
		v.Size(),w.Size());
    throw Oxs_Ext::Error(this,buf);
  }
  Nb_Xpfloat dotsum_xp = 0.0;
  for(UINT4m i=0;i<vecsize;i++) {
    dotsum_xp.Accum(v[i]*w[i]);
  }
  return dotsum_xp.GetValue();
}

void Oxs_LagrangeEvolve::InterchangeCacheAndCurrentData()
{
  UINT4m temp_id = current_id;
  current_id = cache_id;
  cache_id   = temp_id;

  REAL8m temp_L = current_L;
  current_L = cache_L;
  cache_L = temp_L;

  current_x.Swap(cache_x);
  current_energy.Swap(cache_energy);
  current_mxHxm.Swap(cache_mxHxm);
}

BOOL Oxs_LagrangeEvolve::Step(const Oxs_MinDriver* driver,
                        Oxs_ConstKey<Oxs_SimState> current_state_key,
                        const Oxs_DriverStepInfo& /* step_info */,
                        Oxs_Key<Oxs_SimState>& next_state_key)
{
  BOOL step_good=0;
  const Oxs_SimState* cstate = current_state_key.GetPtr();

/**/
#ifndef NDEBUG
static int max_line_min_count=0;
static int line_min_count=0;
#endif // NDEBUG
/**/

  // Initialize next_state structure
  driver->FillState(current_state_key.GetReadReference(),
		    next_state_key.GetWriteReference());

  // First check for stage change
  if(cstate->stage_number != working_stage_number) {
    StageInit(); // In particular, sets lambda_valid, ..., false.
    working_stage_number=cstate->stage_number;
  }

  // Act I: Set lambda
  if(!lambda_valid) {
    if(current_id ==0 || current_id != cstate->Id()) {
      UpdateCurrentData(*cstate);  // Update extra current
      // state data.  Most of the time this should already
      // be set.
    }
    FillLambdaFromCurrentData(*cstate);
    lambda_valid=1;
    base_valid=0;
    base_use_conjugate_direction = 0;
    bracket_valid=0;
    bracket_start_step *= 0.5; // More conservative is
    /// bracket_start_step = minimum_bracket_step;
    /// Less conservative leaves bracket_start_step unmodified.
    lambda_CG_cycle_count=0;
    lambda_CG_cycle_max_L_drop=0.;
    ++lambda_cycle_count;
  }

  // Act II: Set base point
  if(!base_valid) {
#ifndef NDEBUG
line_min_count=0; /**/
#endif // NDEBUG
    if(current_id ==0 || current_id != cstate->Id()) {
      UpdateCurrentData(*cstate);  // Update extra current
      // state data.  Most of the time this should already
      // be set.
    }
    REAL8m Lp = 0;
    if(!base_use_conjugate_direction) {
      //  Act IIa: Set base direction to gradient direction
      base_dL_normsq = 
	Calculate_dL(cstate->mesh,*(cstate->Ms),
		     weighted_rho,current_mxHxm,
		     lambda,current_x,base_direction);
      base_direction *= -1.0;
      Lp = -1*base_dL_normsq;
    } else {
      //  Act IIb: Set base direction in conjugate direction
      // Direction[n] = -dL[n] + gamma*direction[n-1],
      // where 'n' is the cycle count, and
      //            gamma = dL[n]^2/dL[n-1]^2
      // (Fletcher-Reeves formulation)
      REAL8m new_dL_normsq =
	Calculate_dL(cstate->mesh,*(cstate->Ms),
		     weighted_rho,current_mxHxm,
		     lambda,current_x,scratch_field);
      REAL8m gamma = 0.;
      if(base_dL_normsq>0.) gamma = new_dL_normsq/base_dL_normsq;
      base_dL_normsq = new_dL_normsq;
      base_direction *= gamma;
      base_direction -= scratch_field;
      Lp = VecDot(base_direction,scratch_field);
      if(Lp >= -1 * CG_collapse_guard * base_dL_normsq) {
	// Subspace collapsed; Lp might even be slightly
	// positive due to round-off.  Punt; fall back
	// to gradient direction.
	base_direction.Swap(scratch_field);
	base_direction *= -1.0;
	Lp = -1*base_dL_normsq;
      }
    }
    base_direction_max_mag = FindMaxMag(base_direction);
    base_x = current_x;            // Copy current data into base area.
    base_energy = current_energy;
    base_id=current_id;
    base_L=current_L;
    base_valid=1;
    base_use_conjugate_direction=1;
    bracket_valid=0;
    leftpt.offset = midpt.offset = rightpt.offset = 0.;
    leftpt.rel_L  = midpt.rel_L  = rightpt.rel_L  = 0.;
    leftpt.Lp     = midpt.Lp     = rightpt.Lp     = Lp;
    last_minimum_reduction_ratio=1;  // Reset
    next_to_last_minimum_reduction_ratio=1;
    ++direction_cycle_count;
    ++lambda_CG_cycle_count;
  }

  // Act III: Take step
  REAL8m orig_min_rel_L = OC_MIN(OC_MIN(leftpt.rel_L,midpt.rel_L),
				 rightpt.rel_L);
  REAL8m minima_uncertainty=0;
  BOOL right_minimum = 1; // If true, minimum is in righthand interval
  if(!bracket_valid) {
    // Act IIIa: Bracket minimum along base direction
    BOOL bracket_error;
    ++bracket_cycle_count;
    bracket_valid =
      FindBracketStep(cstate,next_state_key,bracket_error);
    if(bracket_error) {
      // Minimum not bracketed in allowed range
      base_valid=0;
      base_use_conjugate_direction=0;
      bracket_valid=0;
      bracket_start_step = maximum_bracket_step;
    }
    minima_uncertainty = rightpt.offset - midpt.offset;
  } else {
    // Act IIIb: Find minimum between brackets along base direction
    ++linemin_cycle_count;
    TakeLineMinimumStep(cstate,next_state_key);
    REAL8m leftspan = midpt.offset - leftpt.offset;
    if(leftspan>0. && (midpt.Lp>=0. || midpt.rel_L>=leftpt.rel_L)) {
      // Minimum between leftpt and midpt
      right_minimum=0;
      minima_uncertainty = leftspan;
    } else {
      // Minimum between midpt and rightpt
      minima_uncertainty = rightpt.offset - midpt.offset;
    }
/**/
#ifndef NDEBUG
++line_min_count;
if(line_min_count>max_line_min_count) {
  max_line_min_count=line_min_count;
}
fprintf(stderr,"Line min count: %d/%d\n",line_min_count,max_line_min_count);
#endif // NDEBUG
/**/
  }

#ifdef INSTRUMENT
  if (bracket_valid) {
    static int test_count=0;
    REAL8m span = rightpt.offset - leftpt.offset;
    if(test_count%25==0 && span>0) {
      fprintf(stderr,">test_count=%d, midpt=%.3f\n",test_count,
              static_cast<double>((midpt.offset-leftpt.offset)/span));
      REAL8m Lscale = fabs(rightpt.rel_L - leftpt.rel_L);
      REAL8m Lpscale = fabs(rightpt.Lp+leftpt.Lp)/2.0;
      if(Lscale==0) {
	Lscale = Lpscale * span;
      } else {
	Lpscale = Lscale / span;
      }
      fprintf(stderr," %3.1f %8.4f %8.4f\n",0.0,
	      0.,static_cast<double>(leftpt.Lp/Lpscale));
      Bracket testpt;
      for(int i=1;i<10;i++) {
	REAL8m offset = leftpt.offset + (i/10.)*span;
	FillBracket(offset,cstate,next_state_key,testpt);
	fprintf(stderr," %3.1f %8.4f %8.4f\n",i/10.,
		static_cast<double>((testpt.rel_L-leftpt.rel_L)/Lscale),
		static_cast<double>(testpt.Lp/Lpscale));
      }
      fprintf(stderr," %3.1f %8.4f %8.4f\n",1.0,
	      static_cast<double>((rightpt.rel_L-leftpt.rel_L)/Lscale),
	      static_cast<double>(rightpt.Lp/Lpscale));
    }
    ++test_count;
  }
#endif // INSTRUMENT

  const Oxs_SimState& nextstate
    = next_state_key.GetDepReference();

  // Check norm drift
  REAL8m normsq_error=0.;
  nextstate.GetDerivedData("Max norm error",normsq_error);
  if(normsq_error>normsq_guard) {
    // Reject step, and restart with new lambda, new spins
    lambda_valid=0;
    return 0;
  }

  REAL8m new_min_rel_L = OC_MIN(OC_MIN(leftpt.rel_L,midpt.rel_L),
				rightpt.rel_L);
  if(new_min_rel_L<orig_min_rel_L) {
    // Accept next_state as valid step
    InterchangeCacheAndCurrentData();
    step_good=1;
  }

  BOOL line_min_found=0;
#ifdef OLD_CODE
  if(rightpt.offset-midpt.offset
     <= line_minimum_relwidth*rightpt.offset) {
    line_min_found=1;
  } else if(bracket_valid && rightpt.Lp<0.) {
    // In this case, rightpt.rel_L should be larger than midpt.rel_L
    if(rightpt.rel_L<=midpt.rel_L) {
      line_min_found=1; // Intended to cover equality case, which can
      /// arise due to roundoff.
    } else {
      REAL8m expected_drop = -midpt.Lp*(rightpt.offset-midpt.offset);
      /// NB: midpt.Lp _should_ be negative
      if(expected_drop <= 
	 REAL8_EPSILON*(fabs(midpt.rel_L)+fabs(rightpt.rel_L))) {
	line_min_found=1;
      }
    }
  }
#else
  if(bracket_valid) {
    REAL8m rel_err = OC_MAX(line_minimum_relwidth,REAL8_EPSILON);
    REAL8m uncert1 = rel_err * base_direction_max_mag;
    REAL8m uncert2 = rel_err * midpt.offset;
    if(minima_uncertainty <= OC_MAX(uncert1,uncert2)) {
      line_min_found=1;
    } else if(right_minimum && rightpt.Lp<0) {
      REAL8m expected_drop = -midpt.Lp*(rightpt.offset-midpt.offset);
      if(expected_drop <= REAL8_EPSILON*fabs(midpt.rel_L)) {
	 line_min_found=1;
      }
    } else if(!right_minimum && midpt.Lp<0) {
      REAL8m expected_drop = -leftpt.Lp*(midpt.offset-leftpt.offset);
      if(expected_drop <= REAL8_EPSILON*fabs(leftpt.rel_L)) {
	 line_min_found=1;
      }
    }
  }
#endif // OLD_CODE

  if(line_min_found) {
    // Line minimization complete for this direction
    base_valid=0; bracket_valid=0;

    // Setup bracket_start_step for next direction
    bracket_start_step = midpt.offset*base_direction_max_mag;
    if(bracket_start_step<=0.0) {
      bracket_start_step=minimum_bracket_step;
    } else if(bracket_start_step>maximum_bracket_step) {
      bracket_start_step=maximum_bracket_step;
    }

    REAL8m L_drop = -1*new_min_rel_L;
    if(L_drop>lambda_CG_cycle_max_L_drop) {
      lambda_CG_cycle_max_L_drop = L_drop;
    }
    if(lambda_CG_cycle_count>=lambda_min_CG_cycle_count
       && L_drop < lambda_reset_L_drop_ratio
                  *lambda_CG_cycle_max_L_drop         ) {
      // Adjust lambda on next pass.
      lambda_valid=0;
    }
  }

  return step_good;
}

void Oxs_LagrangeEvolve::UpdateDerivedOutputs(const Oxs_SimState& state)
{ // This routine fills all the Oxs_LagrangeEvolve Oxs_ScalarOutput's to
  // the appropriate value based on the import "state".  Also any of
  // Oxs_VectorOutput's that have CacheRequest enabled are filled. 
  max_mxHxm_output.cache.state_id
    = delta_E_output.cache.state_id
    = total_energy_output.cache.state_id
    = energy_calc_count_output.cache.state_id
    = L_output.cache.state_id
    = lambda_cycle_count_output.cache.state_id
    = direction_cycle_count_output.cache.state_id
    = bracket_cycle_count_output.cache.state_id
    = linemin_cycle_count_output.cache.state_id
    = 0;  // Mark change in progress

  REAL8m last_energy;
  if(!state.GetDerivedData("Max mxHxm",max_mxHxm_output.cache.value) ||
     !state.GetDerivedData("Last energy",last_energy) ||
     !state.GetDerivedData("Total energy",
			   total_energy_output.cache.value) ||
     !state.GetDerivedData("Energy calc count",
			   energy_calc_count_output.cache.value) ||
     !state.GetDerivedData("Augmented Lagrangian",
			   L_output.cache.value) ||
     !state.GetDerivedData("Lambda cycle count",
			   lambda_cycle_count_output.cache.value) ||
     !state.GetDerivedData("Direction cycle count",
			   direction_cycle_count_output.cache.value) ||
     !state.GetDerivedData("Bracket cycle count",
			   bracket_cycle_count_output.cache.value) ||
     !state.GetDerivedData("Linemin cycle count",
			   linemin_cycle_count_output.cache.value)) {
    // Missing at least some data
    Oxs_MeshValue<REAL8m> dummy_energy;
    Oxs_MeshValue<ThreeVector> dummy_field;
    GetEnergyAndmxHxm(state,dummy_energy,dummy_field);
    if(!state.GetDerivedData("Max mxHxm",max_mxHxm_output.cache.value) ||
       !state.GetDerivedData("Last energy",last_energy) ||
       !state.GetDerivedData("Total energy",
			     total_energy_output.cache.value) ||
       !state.GetDerivedData("Energy calc count",
			     energy_calc_count_output.cache.value) ||
       !state.GetDerivedData("Augmented Lagrangian",
			     L_output.cache.value) ||
       !state.GetDerivedData("Lambda cycle count",
			     lambda_cycle_count_output.cache.value) ||
       !state.GetDerivedData("Direction cycle count",
			     direction_cycle_count_output.cache.value) ||
       !state.GetDerivedData("Bracket cycle count",
			     bracket_cycle_count_output.cache.value) ||
       !state.GetDerivedData("Linemin cycle count",
			     linemin_cycle_count_output.cache.value)) {
      throw Oxs_Ext::Error(this,"Missing output data."
			   " Programming error?");
    }
  }
  delta_E_output.cache.value
    = total_energy_output.cache.value - last_energy;

  max_mxHxm_output.cache.state_id
    = delta_E_output.cache.state_id
    = total_energy_output.cache.state_id
    = energy_calc_count_output.cache.state_id
    = L_output.cache.state_id
    = lambda_cycle_count_output.cache.state_id
    = direction_cycle_count_output.cache.state_id
    = bracket_cycle_count_output.cache.state_id
    = linemin_cycle_count_output.cache.state_id
    = state.Id();
}
