/* FILE: modllg_eulerevolve.h          -*-Mode: c++-*-
 *
 * Concrete evolver class, using simple forward Euler steps.
 * This version implements a modified LLG equation with |m|
 * control.
 *
 */

#ifndef _OXS_MODLLG_EULEREVOLVE
#define _OXS_MODLLG_EULEREVOLVE

#include "timeevolver.h"
#include "key.h"
#include "output.h"

/* End includes */

class Oxs_ModLLG_EulerEvolve:public Oxs_TimeEvolver {
private:
  // Base step size control parameters
  REAL8m min_timestep;           // Seconds
  REAL8m max_timestep;           // Seconds

  // Error-based step size control parameters.  Each may be disabled
  // by setting to -1.  There is an additional step size control that
  // insures that energy is monotonically non-increasing (up to
  // estimated rounding error).
  REAL8m allowed_error_rate;  // Step size is adjusted so
  /// that the estimated maximum error (across all spins) divided
  /// by the step size is smaller than this value.  The units
  /// internally are radians per second, converted from the value
  /// specified in the input MIF file, which is in deg/sec.

  REAL8m allowed_absolute_step_error; // Similar to allowed_error_rate,
  /// but without the step size adjustment.  Internal units are
  /// radians; MIF input units are degrees.

  REAL8m allowed_relative_step_error; // Step size is adjusted so that
  /// the estimated maximum error (across all spins) divided by
  /// [maximum dm/dt (across all spins) * step size] is smaller than
  /// this value.  This value is non-dimensional, representing the
  /// allowed relative (proportional) error, presumably in (0,1).

  REAL8m step_headroom; // The 3 control parameters above can be
  /// used to estimate the step size that would just fit the control
  /// requirements.  Because this is only an estimate, if the step size
  /// is actually set to that value there is a good chance that the
  /// requirement will not be met.  So instead, we leave some headroom
  /// by setting the step size to the computed value multiplied by
  /// step_headroom.  This is a non-dimensional quantity, which should
  /// be in the range (0,1).

  // The following evolution constants are uniform for now.  These
  // should be changed to arrays in the future.
  REAL8m gamma;  // Landau-Lifschitz gyromagnetic ratio
  REAL8m alpha;  // Landau-Lifschitz damping coef
  BOOL do_precess;  // If false, then do pure damping

  // The next timestep is based on the error from the last step.  If
  // there is no last step (either because this is the first step,
  // or because the last state handled by this routine is different
  // from the incoming current_state), then timestep is calculated
  // so that max_dm_dt * timestep = start_dm.
  REAL8m start_dm;

  REAL8m m_mag_adj;    // |m| adjustment, proportional weight in LLG ODE.
  REAL8m m_mag_ie_adj; // "", integrated error weight in LLG ODE.
  REAL8m m_mag_ie_freq; // "", integrated error frequency in LLG ODE.
  /// m_mag_ie_freq is computed from the mag_ie_period input.


  REAL8m allowed_mnorm_deviation; // |m| hard limit.
  enum ResizeType { RT_INVALID, RT_NONE, RT_NORM, RT_H, RT_H_AVE,
                    RT_DMDT_AVE, RT_DMDT_H, RT_H_MOD };
  ResizeType resize_norm,resize_differential;

  // Data cached from last state
  UINT4m energy_state_id;
  Oxs_MeshValue<REAL8m> energy;
  Oxs_MeshValue<ThreeVector> Heff;
  REAL8m next_timestep;
  Oxs_MeshValue<REAL8m> integrated_norm_error;

  // Outputs
  void UpdateDerivedOutputs(const Oxs_SimState&);
  Oxs_ScalarOutput<Oxs_ModLLG_EulerEvolve> max_dm_dt_output;
  Oxs_ScalarOutput<Oxs_ModLLG_EulerEvolve> min_m_output;
  Oxs_ScalarOutput<Oxs_ModLLG_EulerEvolve> max_m_output;
  Oxs_ScalarOutput<Oxs_ModLLG_EulerEvolve> max_mdev_output;
  Oxs_ScalarOutput<Oxs_ModLLG_EulerEvolve> dE_dt_output;
  Oxs_ScalarOutput<Oxs_ModLLG_EulerEvolve> delta_E_output;
  Oxs_VectorFieldOutput<Oxs_ModLLG_EulerEvolve> dm_dt_output;
  Oxs_VectorFieldOutput<Oxs_ModLLG_EulerEvolve> mxH_output;

  // Scratch space
  Oxs_MeshValue<REAL8m> new_energy;
  Oxs_MeshValue<ThreeVector> new_dm_dt;
  Oxs_MeshValue<REAL8m> new_integrated_norm_error;

  void Calculate_dm_dt
  (const Oxs_Mesh& mesh_,
   const Oxs_MeshValue<REAL8m>& Ms_,
   const Oxs_MeshValue<ThreeVector>& mxH_,
   const Oxs_MeshValue<ThreeVector>* old_dm_dt_ptr_,
   const Oxs_MeshValue<ThreeVector>& spin_,
   REAL8m pE_pt_,REAL8m timestep_,
   Oxs_MeshValue<ThreeVector>& dm_dt_,
   REAL8m& max_dm_dt_,REAL8m& dE_dt_,REAL8m& min_timestep_);
  /// Imports: mesh_, Ms_, mxH_, spin_, pE_pt, new_stage_time
  /// Exports: dm_dt_, max_dm_dt_, dE_dt_, min_timestep_,
  ///          min_m_, max_m_

public:
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.
  virtual BOOL Init();
  Oxs_ModLLG_EulerEvolve(const char* name,     // Child instance id
		 Oxs_Director* newdtr, // App director
		 const char* argstr);  // MIF input block parameters
  virtual ~Oxs_ModLLG_EulerEvolve();

  virtual  BOOL
  Step(const Oxs_TimeDriver* driver,
       Oxs_ConstKey<Oxs_SimState> current_state,
       const Oxs_DriverStepInfo& step_info,
       Oxs_Key<Oxs_SimState>& next_state);
  // Returns true if step was successful, false if
  // unable to step as requested.
};

#endif // _OXS_MODLLG_EULEREVOLVE
