/* FILE: modllg_rungekuttaevolve.h          -*-Mode: c++-*-
 *
 * Concrete evolver class, using Runge-Kutta steps.
 * This version implements a modified LLG equation with |m|
 * control.
 *
 */

#ifndef _OXS_MODLLG_RUNGEKUTTAEVOLVE
#define _OXS_MODLLG_RUNGEKUTTAEVOLVE

#include "timeevolver.h"
#include "key.h"
#include "output.h"

/* End includes */
class Oxs_ModLLG_RungeKuttaEvolve:public Oxs_TimeEvolver {
private:
  // Base step size control parameters
  REAL8m min_timestep;           // Seconds
  REAL8m max_timestep;           // Seconds

  const REAL8m max_step_decrease;        // Safety size adjusment
  const REAL8m max_step_increase_limit;  // bounds.
  const REAL8m max_step_increase_adj_ratio;
  REAL8m max_step_increase;
  /// NOTE: These bounds do not include step_headroom, which
  /// is applied at the end.

  // Error-based step size control parameters.  Each may be disabled
  // by setting to -1.  There is an additional step size control that
  // insures that energy is monotonically non-increasing (up to
  // estimated rounding error).
  REAL8m allowed_error_rate;  // Step size is adjusted so
  /// that the estimated maximum error (across all spins) divided
  /// by the step size is smaller than this value.  The units
  /// internally are radians per second, converted from the value
  /// specified in the input MIF file, which is in deg/sec.

  REAL8m allowed_absolute_step_error; // Similar to allowed_error_rate,
  /// but without the step size adjustment.  Internal units are
  /// radians; MIF input units are degrees.

  REAL8m allowed_relative_step_error; // Step size is adjusted so that
  /// the estimated maximum error (across all spins) divided by
  /// [maximum dm/dt (across all spins) * step size] is smaller than
  /// this value.  This value is non-dimensional, representing the
  /// allowed relative (proportional) error, presumably in (0,1).

  REAL8m expected_energy_precision; // Expected relative energy
  /// precision.

  REAL8m reject_goal,reject_ratio;
  REAL8m min_step_headroom,max_step_headroom;
  REAL8m step_headroom; // Safety margin used in step size adjustment

  // The following evolution constants are uniform for now.  These
  // should be changed to arrays in the future.
  REAL8m gamma;  // Landau-Lifschitz gyromagnetic ratio
  REAL8m alpha;  // Landau-Lifschitz damping coef
  BOOL do_precess;  // If false, then do pure damping

  // The next timestep is based on the error from the last step.  If
  // there is no last step (either because this is the first step,
  // or because the last state handled by this routine is different
  // from the incoming current_state), then timestep is calculated
  // so that max_dm_dt * timestep = start_dm.
  REAL8m start_dm;


  REAL8m m_mag_adj;    // |m| adjustment, proportional weight in LLG ODE.
  REAL8m m_mag_ie_adj; // "", integrated error weight in LLG ODE.
  REAL8m m_mag_ie_freq; // "", integrated error frequency in LLG ODE.
  /// m_mag_ie_freq is computed from the mag_ie_period input.
  REAL8m m_mag_de_adj;

  REAL8m m_mag_width;


  REAL8m allowed_mnorm_deviation; // |m| hard limit.
  enum ResizeType { RT_INVALID, RT_NONE, RT_NORM, RT_H, RT_H_AVE,
                    RT_DMDT_AVE, RT_DMDT_H, RT_H_MOD };
  ResizeType resize_norm,resize_differential;

  BOOL ResizeSpins(ResizeType style,
                   const Oxs_MeshValue<ThreeVector>& spin1,
                   Oxs_MeshValue<ThreeVector>& spin2,
                   const Oxs_MeshValue<ThreeVector>& H1,
                   const Oxs_MeshValue<ThreeVector>& H2,
                   const Oxs_MeshValue<ThreeVector>& dm_dt1,
                   const Oxs_MeshValue<ThreeVector>& dm_dt2);

  // Convenience functions
  REAL8m ComputeMaxNormError(REAL8m min_norm,REAL8m max_norm) const;
  REAL8m ComputeMaxNormError(const Oxs_MeshValue<ThreeVector>& spin) const;

  // Stepsize control for first step of each stage after the first.
  // Choices are to use start_dm, use continuation from end of
  // previous stage, or to automatically select between the two
  // methods depending on whether or not the energy appears to
  // be continuous across the stage boundary.
  enum StageInitStepControl { SISC_INVALID, SISC_START_DM,
			      SISC_CONTINUOUS, SISC_AUTO };
  StageInitStepControl stage_init_step_control;

  // Data cached from last state
  UINT4m energy_state_id;
  Oxs_MeshValue<REAL8m> energy;
  Oxs_MeshValue<ThreeVector> Heff;
  REAL8m next_timestep;
  Oxs_MeshValue<REAL8m> integrated_norm_error;
  Oxs_MeshValue<REAL8m> last_norm_error;

  // Outputs
  void UpdateDerivedOutputs(const Oxs_SimState&);
  Oxs_ScalarOutput<Oxs_ModLLG_RungeKuttaEvolve> max_dm_dt_output;
  Oxs_ScalarOutput<Oxs_ModLLG_RungeKuttaEvolve> min_m_output;
  Oxs_ScalarOutput<Oxs_ModLLG_RungeKuttaEvolve> max_m_output;
  Oxs_ScalarOutput<Oxs_ModLLG_RungeKuttaEvolve> max_mdev_output;
  Oxs_ScalarOutput<Oxs_ModLLG_RungeKuttaEvolve> dE_dt_output;
  Oxs_ScalarOutput<Oxs_ModLLG_RungeKuttaEvolve> delta_E_output;
  Oxs_VectorFieldOutput<Oxs_ModLLG_RungeKuttaEvolve> dm_dt_output;
  Oxs_VectorFieldOutput<Oxs_ModLLG_RungeKuttaEvolve> mxH_output;

  // Scratch space
  Oxs_MeshValue<REAL8m> temp_energy;
  Oxs_MeshValue<ThreeVector> tmpHeff;
  Oxs_MeshValue<ThreeVector> vtmpA;
  Oxs_MeshValue<ThreeVector> vtmpB;
  Oxs_MeshValue<ThreeVector> vtmpC;
  Oxs_MeshValue<ThreeVector> vtmpD;
  Oxs_MeshValue<REAL8m> new_integrated_norm_error;
  Oxs_MeshValue<REAL8m> new_norm_error;

  // Utility functions
  void CheckCache(const Oxs_SimState& cstate);

  void AdjustState(REAL8m hstep,
		   REAL8m mstep,
		   const Oxs_SimState& old_state,
		   const Oxs_MeshValue<ThreeVector>& dm_dt,
		   Oxs_SimState& new_state,
		   REAL8m& norm_error) const;
  // Export new state has time index from old_state + h,
  // and spins from old state + mstep*dm_dt and re-normalized.

  void UpdateTimeFields(const Oxs_SimState& cstate,
			Oxs_SimState& nstate,
			REAL8m stepsize) const;

  void NegotiateTimeStep(const Oxs_TimeDriver* driver,
			 const Oxs_SimState&  cstate,
			 Oxs_SimState& nstate,
			 REAL8m stepsize,
			 BOOL use_start_dm,
			 BOOL& forcestep,
			 BOOL& driver_set_step) const;

  BOOL CheckError(REAL8m global_error_order,REAL8m error,
		  REAL8m stepsize,REAL8m reference_stepsize,
		  REAL8m max_dm_dt,REAL8m& new_stepsize);
  /// Returns 1 if step is good, 0 if error is too large.
  /// Export new_stepsize is set to suggested stepsize
  /// for next step.

  REAL8m MaxDiff(const Oxs_MeshValue<ThreeVector>& vecA,
		 const Oxs_MeshValue<ThreeVector>& vecB);
  /// Returns maximum difference between vectors in corresponding
  /// positions in two vector fields.

  void AdjustStepHeadroom(INT4m step_reject);
  /// step_reject should be 0 or 1, reflecting whether the current
  /// step was rejected or not.  This routine updates reject_ratio
  /// and adjusts step_headroom appropriately.

  // Stepper routines:  If routine needs to compute the energy
  // at the new (final) state, then it should store the final
  // energy results in temp_energy, mxH in mxH_output.cache,
  // and dm_dt into the vtmpA scratch array, fill
  // the "Timestep lower bound", "Max dm/dt", "dE/dt", and
  // "pE/pt" derived data fields in nstate, and set the export
  // value new_energy_and_dmdt_computed true.  Otherwise the export
  // value should be set false, and the client routine is responsible
  // for obtaining these values as necessary.  (If possible, it is
  // better to let the client compute these values, because the
  // client may be able to defer computation until it has decided
  // whether or not to keep the step.)

  // ADDENDUM 10-June-2005: new_energy_and_dmdt_computed is set
  // false only by the RK4 routine, which actually fills energy
  // and leaves only dmdt not computed.  This behavior is assumed
  // by the ::Step routine, which does not compute energy, but
  // only dmdt if new_energy_and_dmdt_computed is false.  For the
  // modllg code, Heff is needed to compute dmdt, so we add to
  // the contract that if new_energy_and_dmdt_computed is false
  // then additionally Heff should hold H for the new state.

  // One would like to declare the step functions and pointer
  // to same via typedef's, but the MS VC++ 6.0 (& others?)
  // compiler doesn't handle member function typedef's properly---
  // it produces __cdecl linkage rather than instance member
  // linkage.  Typedef's on pointers to member functions work
  // okay, just not typedef's on member functions themselves.
  // So, instead we use a #define, which is ugly but portable.
#define RKStepFuncSig(NAME) \
  void NAME (                                            \
     REAL8m stepsize,                                    \
     Oxs_ConstKey<Oxs_SimState> current_state,           \
     const Oxs_MeshValue<ThreeVector>& current_dm_dt,    \
     Oxs_Key<Oxs_SimState>& next_state,                  \
     REAL8m& error_estimate,                             \
     REAL8m& global_error_order,                         \
     REAL8m& norm_error,                                 \
     BOOL& new_energy_and_dmdt_computed)

  // Functions that calculate a single RK step
  RKStepFuncSig(TakeRungeKuttaStep2);
  RKStepFuncSig(TakeRungeKuttaStep4);
  RKStepFuncSig(TakeRungeKuttaFehlbergStep54);
  RKStepFuncSig(TakeRungeKuttaFehlbergStep54M);
  RKStepFuncSig(TakeRungeKuttaFehlbergStep54S);

  // Pointer set at runtime during instance initialization
  // to one of the above functions single RK step functions.
  RKStepFuncSig((Oxs_ModLLG_RungeKuttaEvolve::* rkstep_ptr));

  // Utility code used by the TakeRungeKuttaFehlbergStep54* routines.
  enum RKF_SubType { RKF_INVALID, RK547FC, RK547FM, RK547FS };
  void RungeKuttaFehlbergBase54(RKF_SubType method,
			   REAL8m stepsize,
			   Oxs_ConstKey<Oxs_SimState> current_state,
			   const Oxs_MeshValue<ThreeVector>& current_dm_dt,
			   Oxs_Key<Oxs_SimState>& next_state,
			   REAL8m& error_estimate,
			   REAL8m& global_error_order,
			   REAL8m& norm_error,
			   BOOL& new_energy_and_dmdt_computed);

  void Calculate_dm_dt
  (const Oxs_SimState& state_,
   const Oxs_MeshValue<ThreeVector>& mxH_,
   const Oxs_MeshValue<ThreeVector>* old_dm_dt_ptr_,
   REAL8m pE_pt_,REAL8m timestep_,
   Oxs_MeshValue<ThreeVector>& dm_dt_,
   REAL8m& max_dm_dt_,REAL8m& dE_dt_,REAL8m& min_timestep_,
   REAL8m& min_m_,REAL8m& max_m_);
  /// Imports: mesh_, Ms_, mxH_, spin_, pE_pt
  /// Exports: dm_dt_, max_dm_dt_, dE_dt_, min_timestep_
  ///          min_m_, max_m_

  // Declare but leave undefined copy constructor and assignment operator
  Oxs_ModLLG_RungeKuttaEvolve(const Oxs_ModLLG_RungeKuttaEvolve&);
  Oxs_ModLLG_RungeKuttaEvolve& operator=(const Oxs_ModLLG_RungeKuttaEvolve&);

public:
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.
  virtual BOOL Init();
  Oxs_ModLLG_RungeKuttaEvolve(const char* name,     // Child instance id
		       Oxs_Director* newdtr, // App director
		       const char* argstr);  // MIF input block parameters
  virtual ~Oxs_ModLLG_RungeKuttaEvolve();

  virtual  BOOL
  Step(const Oxs_TimeDriver* driver,
       Oxs_ConstKey<Oxs_SimState> current_state,
       const Oxs_DriverStepInfo& step_info,
       Oxs_Key<Oxs_SimState>& next_state);
  // Returns true if step was successful, false if
  // unable to step as requested.
};

#endif // _OXS_MODLLG_RUNGEKUTTAEVOLVE
