/* FILE: semianalyticevolve.h                 -*-Mode: c++-*-
 *
 * Concrete evolver class, using semi-analytic steps.
 *
 */

#ifndef _OXS_SEMIANALYTICEVOLVE
#define _OXS_SEMIANALYTICEVOLVE

#include "timeevolver.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "scalarfield.h"
#include "output.h"

/* End includes */
class Oxs_SemiAnalyticEvolve:public Oxs_TimeEvolver {
private:
  mutable UINT4m mesh_id;     // Used by gamma and alpha meshvalues to
  void UpdateMeshArrays(const Oxs_Mesh*);   /// track changes in mesh.

  // Base step size control parameters
  REAL8m min_timestep;           // Seconds
  REAL8m max_timestep;           // Seconds

  const REAL8m max_step_decrease;        // Safety size adjusment
  const REAL8m max_step_increase_limit;  // bounds.
  const REAL8m max_step_increase_adj_ratio;
  REAL8m max_step_increase;
  /// NOTE: These bounds do not include step_headroom, which
  /// is applied at the end.

  // Error-based step size control parameters.  Each may be disabled
  // by setting to -1.  There is an additional step size control that
  // insures that energy is monotonically non-increasing (up to
  // estimated rounding error).
  REAL8m allowed_error_rate;  // Step size is adjusted so
  /// that the estimated maximum error (across all spins) divided
  /// by the step size is smaller than this value.  The units
  /// internally are radians per second, converted from the value
  /// specified in the input MIF file, which is in deg/sec.

  REAL8m allowed_absolute_step_error; // Similar to allowed_error_rate,
  /// but without the step size adjustment.  Internal units are
  /// radians; MIF input units are degrees.

  REAL8m allowed_relative_step_error; // Step size is adjusted so that
  /// the estimated maximum error (across all spins) divided by
  /// [maximum dm/dt (across all spins) * step size] is smaller than
  /// this value.  This value is non-dimensional, representing the
  /// allowed relative (proportional) error, presumably in (0,1).

  REAL8m expected_energy_precision; // Expected relative energy
  /// precision.

  REAL8m reject_goal,reject_ratio;
  REAL8m min_step_headroom,max_step_headroom;
  REAL8m step_headroom; // Safety margin used in step size adjustment

  // Stepsize control error debugging
  BOOL debug_step_error;
  String debug_step_error_basename;
  int debug_step_error_start;
  int debug_step_error_stop;
  int debug_step_error_iteration_count;
  int debug_step_error_iteration_subcount;

  // Spatially variable Landau-Lifschitz-Gilbert gyromagnetic ratio
  // and damping coefficients.
  BOOL do_precess;  // If false, then do pure damping
  BOOL allow_signed_gamma; // If false, then force gamma negative
  enum GammaStyle { GS_INVALID, GS_LL, GS_G }; // Landau-Lifshitz or Gilbert
  GammaStyle gamma_style;
  Oxs_OwnedPointer<Oxs_ScalarField> gamma_init;
  mutable Oxs_MeshValue<REAL8m> gamma;

  Oxs_OwnedPointer<Oxs_ScalarField> alpha_init;
  mutable Oxs_MeshValue<REAL8m> alpha;

  // The next timestep is based on the error from the last step.  If
  // there is no last step (either because this is the first step,
  // or because the last state handled by this routine is different
  // from the incoming current_state), then timestep is calculated
  // so that max_dm_dt * timestep = start_dm.
  REAL8m start_dm;

  // Stepsize control for first step of each stage after the first.
  // Choices are to use start_dm, use continuation from end of
  // previous stage, or to automatically select between the two
  // methods depending on whether or not the energy appears to
  // be continuous across the stage boundary.
  enum StageInitStepControl { SISC_INVALID, SISC_START_DM,
			      SISC_CONTINUOUS, SISC_AUTO };
  StageInitStepControl stage_init_step_control;

  // Data cached from last state
  UINT4m energy_state_id;
  Oxs_MeshValue<REAL8m> energy;
  Oxs_MeshValue<ThreeVector> H_cached;
  REAL8m next_timestep;
  REAL8m old_timestep_cached;
  Oxs_MeshValue<ThreeVector> m_old_cached;
  Oxs_MeshValue<ThreeVector> mxH_old_cached;
  Oxs_MeshValue<ThreeVector> mxHxm_old_cached;
  BOOL mxH_old_cached_is_valid; // True if old_timestep_cached,
  /// m_old_cached, mxH_old_cached, and mxHxm_old_cached are *all* valid.


  // Outputs
  void UpdateDerivedOutputs(const Oxs_SimState&);
  Oxs_ScalarOutput<Oxs_SemiAnalyticEvolve> max_dm_dt_output;
  Oxs_ScalarOutput<Oxs_SemiAnalyticEvolve> dE_dt_output;
  Oxs_ScalarOutput<Oxs_SemiAnalyticEvolve> delta_E_output;
  Oxs_VectorFieldOutput<Oxs_SemiAnalyticEvolve> mxH_output;

  // Scratch space
  Oxs_MeshValue<REAL8m> temp_energy;
  Oxs_MeshValue<ThreeVector> vtmpA;
  Oxs_MeshValue<ThreeVector> vtmpB;
  Oxs_MeshValue<ThreeVector> vtmpC;
  Oxs_MeshValue<ThreeVector> vtmpD;
  Oxs_MeshValue<ThreeVector> vtmpE;
  Oxs_MeshValue<ThreeVector> vtmpF;

  // Utility functions
  void CheckCache(const Oxs_SimState& cstate);

  void ZeroFixedSpinFields(Oxs_MeshValue<ThreeVector>* mxH,
                           Oxs_MeshValue<ThreeVector>* H) const;

  void UpdateTimeFields(const Oxs_SimState& cstate,
			Oxs_SimState& nstate,
			REAL8m stepsize) const;

  void NegotiateTimeStep(const Oxs_TimeDriver* driver,
			 const Oxs_SimState&  cstate,
			 Oxs_SimState& nstate,
			 REAL8m stepsize,
			 BOOL use_start_dm,
			 BOOL& forcestep,
			 BOOL& driver_set_step) const;

  BOOL CheckError(REAL8m global_error_order,REAL8m error,
		  REAL8m stepsize,REAL8m reference_stepsize,
		  REAL8m max_dm_dt,REAL8m& new_stepsize);
  /// Returns 1 if step is good, 0 if error is too large.
  /// Export new_stepsize is set to suggested stepsize
  /// for next step.

  REAL8m MaxDiff(const Oxs_MeshValue<ThreeVector>& vecA,
		 const Oxs_MeshValue<ThreeVector>& vecB);
  /// Returns maximum difference between vectors in corresponding
  /// positions in two vector fields.

  void AdjustStepHeadroom(INT4m step_reject);
  /// step_reject should be 0 or 1, reflecting whether the current
  /// step was rejected or not.  This routine updates reject_ratio
  /// and adjusts step_headroom appropriately.

  void AdjustState(REAL8m hstep,
		   REAL8m mstep,
		   const Oxs_SimState& old_state,
		   const Oxs_MeshValue<ThreeVector>& H,
		   Oxs_SimState& new_state) const;
  /// Export new state has time index from old_state + hstep,
  /// and spins from old state advanced by time mstep.
  /// NOTE: H should be zeroed at fixed spin sites before
  /// being sent to this function.  (In other words, this
  /// function makes no fixed spin consideration.)

  void AdjustState(REAL8m hstep,
		   REAL8m mstep,
		   const Oxs_SimState& old_state,
		   const Oxs_MeshValue<ThreeVector>& H,
		   const Oxs_MeshValue<ThreeVector>& mxH,
		   Oxs_SimState& new_state) const;
  /// Same as preceding AdjustState, except uses an additional
  /// import (mxH) in an attempt to improve numerics.

  void AxisProject(const Oxs_MeshValue<ThreeVector>& m,
                   const Oxs_MeshValue<ThreeVector>& H,
                   const Oxs_MeshValue<ThreeVector>& axis,
                   Oxs_MeshValue<ThreeVector>& Hproj) const;
  // Export Hproj is a projection of H onto axis (i.e., Hproj || axis)
  // such that m x H == m x Hproj

  void AxisCorrectedAdjustState(REAL8m hstep,
                                REAL8m mstep,
                                const Oxs_SimState& current_state,
                                const Oxs_MeshValue<ThreeVector>& mxH_old,
                                const Oxs_MeshValue<ThreeVector>& mxH_current,
                                Oxs_SimState& new_state) const;
  // Similar to AdjustState, but takes two torques as input and advances
  // spins about the axis Cross(mxH_old,mxH_current).

  void SemiAnalyticSubstep2
  (REAL8m stepsize,
   Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
   const Oxs_MeshValue<ThreeVector>& H,          // Field at cstate
   const Oxs_SimState* base_state,               // Used for updating state
   REAL8m effective_stepsize,                   /// time fields
   Oxs_Key<Oxs_SimState>& next_state_key         // Final state
   );
  // If base_state is null, then time fields are not updated.
  // If base_state is not null, then the time fields in next_state
  // are updated from base_state with a step size of effective_stepsize
  // NOTE: Uses temp_energy and vtmpF as scratch space.

  void SemiAnalyticAxisCorrectedSubstep2
  (REAL8m stepsize,
   Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
   const Oxs_MeshValue<ThreeVector>& H,          // Field at cstate
   const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
   const Oxs_MeshValue<ThreeVector>& Torque_alt, // Alternative torque,
                                     /// used to compute rotation axis
   const Oxs_SimState* base_state,               // Used for updating state
   REAL8m effective_stepsize,                   /// time fields
   Oxs_Key<Oxs_SimState>& next_state_key         // Final state
   );
  // If base_state is null, then time fields are not updated.
  // If base_state is not null, then the time fields in next_state
  // are updated from base_state with a step size of effective_stepsize
  // NOTE: Uses temp_energy, vtmpA, and vtmpB as scratch space.

  void SemiAnalyticAxisCorrectedSubstep2b
  (REAL8m old_stepsize,             // timestep between prev and cstate
   const Oxs_MeshValue<ThreeVector>& m_old,      // Spin at prev state
   const Oxs_MeshValue<ThreeVector>& Torque_old, // Torque at prev state
   REAL8m stepsize,                              // current timestep
   Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
   const Oxs_MeshValue<ThreeVector>& H,          // Field at cstate
   const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
   const Oxs_SimState* base_state,               // Used for updating state
   REAL8m effective_stepsize,                   ///    time fields
   Oxs_Key<Oxs_SimState>& next_state_key         // Final state
   );
  // If base_state is null, then time fields are not updated.
  // If base_state is not null, then the time fields in next_state
  // are updated from base_state with a step size of effective_stepsize
  // NOTE: Uses temp_energy, vtmpA, and vtmpB as scratch space.

  static void ComputeAxisCorrectedH
  (REAL8m stepsize,REAL8m lambda,
   REAL8m gamma,REAL8m rotation_limit,REAL8m stability_adj,
   const ThreeVector& m0, const ThreeVector& T0,
   const ThreeVector& m1, const ThreeVector& T1,
   ThreeVector& acH,ThreeVector& acT0,ThreeVector& acT1);

  void SemiAnalyticAxisCorrectedSubstep2c
  (REAL8m old_stepsize,             // timestep between prev and cstate
   const Oxs_MeshValue<ThreeVector>& m_old,      // Spin at prev state
   const Oxs_MeshValue<ThreeVector>& Torque_old, // Torque at prev state
   REAL8m stepsize,                              // current timestep
   Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
   const Oxs_MeshValue<ThreeVector>& H,          // Field at cstate
   const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
   const Oxs_SimState* base_state,               // Used for updating state
   REAL8m effective_stepsize,                   ///    time fields
   Oxs_Key<Oxs_SimState>& next_state_key,        // Final state
   REAL8m rotation_limit       // Allowed rotation, in fraction
   /// of a complete revolution, e.g., 0.5 => max half rotation.
   );
  // If base_state is null, then time fields are not updated.
  // If base_state is not null, then the time fields in next_state
  // are updated from base_state with a step size of effective_stepsize
  // NOTE: Uses temp_energy, vtmpA, and vtmpB as scratch space.

  void SemiAnalyticTorqueSubstep2
  (REAL8m stepsize,
   Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
   const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
   const Oxs_SimState* base_state,               // Used for updating state
   REAL8m effective_stepsize,                   /// time fields
   Oxs_Key<Oxs_SimState>& next_state_key         // Final state
   );
  // If base_state is null, then time fields are not updated.
  // If base_state is not null, then the time fields in next_state
  // are updated from base_state with a step size of effective_stepsize
  // NOTE: Uses temp_energy, vtmpA as scratch space.

  void SemiAnalyticTorqueSubstep2b
  (REAL8m old_stepsize,             // timestep between prev and cstate
   const Oxs_MeshValue<ThreeVector>& old_mxHxm,  // mxHxm at prev state
   REAL8m stepsize,                              // Current stepsize
   Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
   const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
   const Oxs_SimState* base_state,               // Used for updating state
   REAL8m effective_stepsize,                   /// time fields
   Oxs_Key<Oxs_SimState>& next_state_key         // Final state
   );


  // Utility code used by the TakeRungeKuttaFehlbergStep54* routines.
  enum RKF_SubType { RKF_INVALID, RK547FC, RK547FM, RK547FS };
  struct RKF54_dataobj {
  public:
    // Rotated coords
    ThreeVector ex,ey,ez;
    // Spherical coords at base point (m1).
    // Theta is angle from ez, phi is angle from ex,
    // in the ex-ey plane.  That is,
    //  <m1,ex> = sin(theta)cos(phi)
    //  <m1,ey> = sin(theta)sin(phi)
    //  <m1,ex> = cos(theta)
    //  m1 x ez = sin(theta)
    // ex is selected so phi1=0, so phi1 does not need to be stored.
    // However, for development purposes, we include phi1 to make it
    // easier to do a comparison with a fixed coordinate system.
    REAL8m theta1;
    REAL8m phi1;

    // dm_dt in spherical coords, at each of the seven points:
    REAL8m dtheta[7];
    REAL8m dphi[7];

    // Constructor
    RKF54_dataobj() : theta1(0), phi1(0) {
      for(int i=0;i<7;++i) dtheta[i] = dphi[i] = 0.0;
    }
  };
  Oxs_MeshValue<RKF54_dataobj> rkf54_data;

  void RungeKuttaFehlbergBase54_Substep
  (REAL8m stepsize,
   const Oxs_SimState* cstate,
   Oxs_Key<Oxs_SimState>& next_state_key,
   const Oxs_MeshValue<ThreeVector>* mxH_ptr,
   Oxs_MeshValue<ThreeVector>* H_ptr,
   int N,      // "N" is 1-based point number
   REAL8m aN,  // Actual timestep is aN*stepsize
   REAL8m* bN, // Array of length N (0-based)
   REAL8m& pE_pt);

  void RungeKuttaFehlbergBase54(RKF_SubType method,
			   REAL8m stepsize,
			   Oxs_ConstKey<Oxs_SimState> current_state,
                           const Oxs_MeshValue<ThreeVector>& mxH,
			   Oxs_Key<Oxs_SimState>& next_state,
			   REAL8m& error_estimate,
			   REAL8m& global_error_order,
			   BOOL& new_energy_and_field_computed);

  // Stepper routines:  If routine needs to compute the energy
  // at the new (final) state, then it should store the final
  // energy results in temp_energy, mxH in vtmpA, H into vtmpB,
  // fill the "Timestep lower bound", "Max dm/dt", "dE/dt", and
  // "pE/pt" derived data fields in nstate, and set the export
  // value new_energy_and_dmdt_computed true.  Otherwise the export
  // value should be set false, and the client routine is responsible
  // for obtaining these values as necessary.  (If possible, it is
  // better to let the client compute these values, because the
  // client may be able to defer computation until it has decided
  // whether or not to keep the step.)

  // One would like to declare the step functions and pointer
  // to same via typedef's, but the MS VC++ 6.0 (& others?)
  // compiler doesn't handle member function typedef's properly---
  // it produces __cdecl linkage rather than instance member
  // linkage.  Typedef's on pointers to member functions work
  // okay, just not typedef's on member functions themselves.
  // So, instead we use a #define, which is ugly but portable.
#define SAStepFuncSig(NAME) \
  void NAME (                                            \
     REAL8m stepsize,                                    \
     Oxs_ConstKey<Oxs_SimState> state_key,               \
     const Oxs_MeshValue<ThreeVector>& H,                \
     const Oxs_MeshValue<ThreeVector>& mxH,              \
     Oxs_Key<Oxs_SimState>& next_state_key,              \
     REAL8m& error_estimate,                             \
     REAL8m& global_error_order,                         \
     BOOL& new_energy_and_dmdt_computed)

  // Functions that calculate a single SA step
  SAStepFuncSig(TakeSemiAnalyticStep1);
  SAStepFuncSig(TakeSemiAnalyticStep2);
  SAStepFuncSig(TakeSemiAnalyticAxisCorrectedStep1);
  SAStepFuncSig(TakeSemiAnalyticAxisCorrectedStep2);
  SAStepFuncSig(TakeSemiAnalyticAxisCorrectedStep2b);
  SAStepFuncSig(TakeSemiAnalyticAxisCorrectedStep2c);
  SAStepFuncSig(TakeSemiAnalyticTorqueStep2);
  SAStepFuncSig(TakeSemiAnalyticTorqueStep2b);
  SAStepFuncSig(TakeRungeKuttaFehlbergStep54M);
  SAStepFuncSig(TakeRungeKuttaFehlbergStep54S);
  SAStepFuncSig(TakeRungeKuttaFehlbergStep54);

  // Pointer set at runtime during instance initialization
  // to one of the above functions single SA step functions.
  SAStepFuncSig((Oxs_SemiAnalyticEvolve::* sastep_ptr));

  void Compute_dm_dt(const ThreeVector& m,
                     const ThreeVector& mxH,
                     const REAL8m gamma,
                     const REAL8m alpha,
                     ThreeVector& dm_dt) const;

  void ComputeMiscValues
  (const Oxs_SimState& state,
   const Oxs_MeshValue<ThreeVector>& mxH,
   REAL8m pE_pt,
   REAL8m& max_dm_dt,REAL8m& dE_dt,REAL8m& min_timestep) const;

  // Declare but leave undefined copy constructor and assignment operator
  Oxs_SemiAnalyticEvolve(const Oxs_SemiAnalyticEvolve&);
  Oxs_SemiAnalyticEvolve& operator=(const Oxs_SemiAnalyticEvolve&);

public:
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.
  virtual BOOL Init();
  Oxs_SemiAnalyticEvolve(const char* name,     // Child instance id
		       Oxs_Director* newdtr, // App director
		       const char* argstr);  // MIF input block parameters
  virtual ~Oxs_SemiAnalyticEvolve();

  virtual  BOOL
  Step(const Oxs_TimeDriver* driver,
       Oxs_ConstKey<Oxs_SimState> current_state,
       const Oxs_DriverStepInfo& step_info,
       Oxs_Key<Oxs_SimState>& next_state);
  // Returns true if step was successful, false if
  // unable to step as requested.
};

#endif // _OXS_SEMIANALYTICEVOLVE
