/* FILE: arraywrapper.h                 -*-Mode: c++-*-
 *
 * Basic array templates --- these are essentially exception-safe
 * wrappers for the standard C++ new[]/delete[] operators.
 *
 * Note: Compare also against the older Array2D and Array3D classes.
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2007-03-22 01:15:20 $
 * Last modified by: $Author: donahue $
 */

#ifndef _NB_ARRAYWRAPPER
#define _NB_ARRAYWRAPPER

#include "oc.h"

/* End includes */

#define NB_ARRAYWRAPPER_ERRBUFSIZE 1024

////////////////////////////////////////////////////////////////////////
// Wrapper template for standard C++ new[]/delete[] memory allocation.
// Bare new's inside class constructors don't provide automatic cleanup
// (i.e., delete[] calls) if an exception is thrown inside the
// constructor.  On the other hand, STL container classes place copy
// constructor and operator= requirements on their templated class.
// This class provides a workaround for both problems.
template<class T> class Nb_ArrayWrapper
{
private:
  UINT4m size;
  T* objs;

  // Disable copy operations by declaring the following
  // without providing definitions.
  Nb_ArrayWrapper(const Nb_ArrayWrapper<T>&);
  Nb_ArrayWrapper<T>& operator=(const Nb_ArrayWrapper<T>&);

public:
  void Free(); // Calls destructors for objs, releases memory,
  /// and sets size to 0.

  void SetSize(UINT4m new_size);  // Throws an exception
  // if size!=0 on entry.  If you want to use this function,
  // call Free() first.

  UINT4m GetSize() const { return size; }

  Nb_ArrayWrapper(UINT4m initial_size=0) : size(0), objs(NULL) {
    SetSize(initial_size);
  }

  ~Nb_ArrayWrapper() { Free(); }

#ifdef NDEBUG
  T& operator[](UINT4m i) { return objs[i]; }
  const T& operator[](UINT4m i) const { return objs[i]; }
  /// NB: No range check!
#else
  T& operator[](UINT4m i);
  const T& operator[](UINT4m i) const;
  /// Ranged checked versions
#endif

};

template<class T>
void Nb_ArrayWrapper<T>::Free()
{
  if(size==0 && objs==NULL) return; // Nothing to do
  if(size==0 || objs==NULL) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ArrayWrapper<T>","Free",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,
                          "Size and objs pointer incompatibility detected."));
  }
  size=0;
  delete[] objs;
  objs=NULL;
}

template<class T>
void Nb_ArrayWrapper<T>::SetSize(UINT4m new_size)
{
  if(size!=0) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ArrayWrapper<T>","SetSize",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,
                          "Array not empty; call Free first."));
  }
  if(new_size==0) return; // Nothing to do.
  objs = new T[new_size];
  if(objs==NULL) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ArrayWrapper<T>","SetSize",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,
                          "Memory allocation error."));
  }
  size=new_size;
}

#ifndef NDEBUG
template<class T>
T& Nb_ArrayWrapper<T>::operator[](UINT4m i)
{
  if(i>=size) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ArrayWrapper<T>","operator[]",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,"Index too big."));
  } 
  return objs[i];
}

template<class T>
const T& Nb_ArrayWrapper<T>::operator[](UINT4m i) const
{
  if(i>=size) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ArrayWrapper<T>",
                          "operator[] const",NB_ARRAYWRAPPER_ERRBUFSIZE,
                          "Index too big."));
  } 
  return objs[i];
}
#endif // NDEBUG

////////////////////////////////////////////////////////////////////////
// 2D array wrapper.  See notes with Nb_ArrayWrapper above.
// NOTE: Unlike the Oxs_RectangularMesh class, Nb_2DArrayWrapper
//  stores values in normal C order, i.e., fast loops access index
//  (i,j) by varying j in the inner loop, i in the outer loop.
template<class T> class Nb_2DArrayWrapper
{
private:
  UINT4m size1,size2;
  T** objs;

  // Disable copy operations by declaring the following
  // without providing definitions.
  Nb_2DArrayWrapper(const Nb_2DArrayWrapper<T>&);
  Nb_2DArrayWrapper<T>& operator=(const Nb_2DArrayWrapper<T>&);

public:
  void Free(); // Calls destructors for objs, releases memory,
  /// and sets size to 0.

  void SetSize(UINT4m new_size1,UINT4m new_size2);  // Throws an
  // exception if either size!=0 on entry.  If you want to use this
  // function, call Free() first.

  void GetSize(UINT4m& mysize1,UINT4m& mysize2) const {
    mysize1=size1; mysize2=size2;
  }

  Nb_2DArrayWrapper(UINT4m isize1=0,UINT4m isize2=0)
    : size1(0), size2(0), objs(NULL) {
    SetSize(isize1,isize2);
  }

  ~Nb_2DArrayWrapper() { Free(); }

#ifdef NDEBUG
  T* operator[](UINT4m i) { return objs[i]; }
  const T* operator[](UINT4m i) const { return objs[i]; }
  T& operator()(UINT4m i,UINT4m j) { return objs[i][j]; }
  const T& operator()(UINT4m i,UINT4m j) const { return objs[i][j]; }
  /// NB: No range check!
#else
  T* operator[](UINT4m i);
  const T* operator[](UINT4m i) const;
  T& operator()(UINT4m i,UINT4m j);
  const T& operator()(UINT4m i,UINT4m j) const;
  /// Ranged checked versions.  IMPORTANT: The 'operator[]' version only
  /// checks the leading index.  To get full range check protection, use
  /// the operator(,) version.
#endif

};

template<class T>
void Nb_2DArrayWrapper<T>::Free()
{
  if(size1==0 && size2==0 && objs==NULL) return; // Nothing to do
  if(size1==0 || size2==0 || objs==NULL) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>","Free",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,
                          "Size and objs pointer incompatibility detected."));
  }
  size1=size2=0;
  delete[] objs[0];
  delete[] objs;
  objs=NULL;
}

template<class T>
void Nb_2DArrayWrapper<T>::SetSize(UINT4m new_size1,UINT4m new_size2)
{
  if(size1!=0 || size2!=0 || objs!=NULL) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>","SetSize",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,
                          "Array not empty; call Free first."));
  }
  if(new_size1==0 || new_size2==0) return; // Nothing to do.
  UINT4m totalsize = new_size1*new_size2;

  // Partial overflow check
  if(totalsize<new_size1 || totalsize<new_size2) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>","SetSize",
                NB_ARRAYWRAPPER_ERRBUFSIZE,
                "Integer overflow detected; size1=%lu, size2=%lu",
                static_cast<unsigned long>(new_size1),
                static_cast<unsigned long>(new_size2)));
  }
  size1=new_size1;
  size2=new_size2;

  // Allocate pointer array
  objs = new T*[size1];
  if(objs==NULL) {
    // Safety check for old compilers that don't throw exception
    // on memory allocation failure.
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>","SetSize",
                     NB_ARRAYWRAPPER_ERRBUFSIZE,"Memory allocation error."));
  }

  // Allocate base array
  objs[0] = new T[totalsize];
  if(objs[0]==NULL) {
    // Safety check for old compilers that don't throw exception
    // on memory allocation failure.
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>","SetSize",
                     NB_ARRAYWRAPPER_ERRBUFSIZE,"Memory allocation error."));
  }

  // Initialize pointers into base area
  for(UINT4m i=1;i<size1;i++) objs[i] = objs[i-1] + size2;
}

#ifndef NDEBUG
template<class T>
T* Nb_2DArrayWrapper<T>::operator[](UINT4m i)
{
  if(i>=size1) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>",
                          "operator[]",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,"Index too big."));
  }
  return objs[i];
}

template<class T>
const T* Nb_2DArrayWrapper<T>::operator[](UINT4m i) const
{
  if(i>=size1) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>",
                          "operator[] const",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,"Index too big."));
  } 
  return objs[i];
}

template<class T>
T& Nb_2DArrayWrapper<T>::operator()(UINT4m i,UINT4m j)
{
  if(i>=size1 || j>=size2) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>",
                          "operator(UINT4m,UINT4m)",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,"Index too big."));
  } 
  return objs[i][j];
}

template<class T>
const T& Nb_2DArrayWrapper<T>::operator()(UINT4m i,UINT4m j) const
{
  if(i>=size1 || j>=size2) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_2DArrayWrapper<T>",
                          "operator(UINT4m,UINT4m) const",
                          NB_ARRAYWRAPPER_ERRBUFSIZE,"Index too big."));
  } 
  return objs[i][j];
}
#endif // NDEBUG

#endif // _NB_ARRAYWRAPPER
